from __future__ import annotations

import stat
import unicodedata
from functools import cached_property, lru_cache
from io import BytesIO
from typing import TYPE_CHECKING, Any
from uuid import UUID

from dissect.fve.crypto import create_cipher
from dissect.util.hash.crc32c import crc32c
from dissect.util.stream import BufferedStream
from dissect.util.ts import from_unix, from_unix_ns

from dissect.apfs.c_apfs import FILESYSTEM_OBJECT_TYPE_MAP, XF_MAP, c_apfs
from dissect.apfs.cursor import Cursor
from dissect.apfs.exception import Error, FileNotFoundError, NotADirectoryError, NotASymlinkError
from dissect.apfs.objects.base import Object
from dissect.apfs.objects.btree import BTree
from dissect.apfs.objects.omap import ObjectMap
from dissect.apfs.stream import DecmpfsStream, FileStream
from dissect.apfs.util import cmp_fs, cmp_fs_dir, cmp_fs_dir_hash, parse_fs_object_key

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from dissect.apfs.objects.btree_node import BTreeNode
    from dissect.apfs.objects.keybag import VolumeKeybag


class FS(Object):
    """APFS Filesystem object, also referred to as the "volume"."""

    __type__ = c_apfs.OBJECT_TYPE_FS
    __struct__ = c_apfs.apfs_superblock
    object: c_apfs.apfs_superblock

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.object.apfs_magic.to_bytes(4, "big") != c_apfs.APFS_MAGIC:
            raise Error(
                "Invalid apfs_superblock magic "
                f"(expected {c_apfs.APFS_MAGIC!r}, got {self.object.apfs_magic.to_bytes(4, 'big')!r})"
            )

        self._cipher = None

        self.inode = lru_cache(4096)(self.inode)
        self._cursor_state = lru_cache(128)(self._cursor_state)

        self.root = self.inode(c_apfs.ROOT_DIR_INO_NUM)

    def __repr__(self) -> str:
        return f"<FS name={self.name!r} uuid={self.uuid}>"

    @cached_property
    def index(self) -> int:
        """The volume index within the container."""
        return self.object.apfs_fs_index

    @cached_property
    def features(self) -> c_apfs.APFS_FEATURE:
        """The features supported by this filesystem."""
        return c_apfs.APFS_FEATURE(self.object.apfs_features)

    @cached_property
    def incompatible_features(self) -> c_apfs.APFS_INCOMPAT:
        """The incompatible features supported by this filesystem."""
        return c_apfs.APFS_INCOMPAT(self.object.apfs_incompatible_features)

    @cached_property
    def is_case_insensitive(self) -> bool:
        """Whether the filesystem is case insensitive."""
        return c_apfs.APFS_INCOMPAT.CASE_INSENSITIVE in self.incompatible_features

    @cached_property
    def is_normalization_insensitive(self) -> bool:
        """Whether the filesystem is normalization insensitive."""
        return c_apfs.APFS_INCOMPAT.NORMALIZATION_INSENSITIVE in self.incompatible_features

    @cached_property
    def is_sealed(self) -> bool:
        """Whether the filesystem is sealed (read-only)."""
        return c_apfs.APFS_INCOMPAT.SEALED_VOLUME in self.incompatible_features

    @cached_property
    def unmount_time(self) -> datetime:
        """The last unmount time of the filesystem."""
        return from_unix(self.object.apfs_unmount_time)

    @cached_property
    def omap(self) -> ObjectMap:
        """The object map for the filesystem."""
        return ObjectMap.from_address(self.container, self.object.apfs_omap_oid)

    @cached_property
    def root_tree(self) -> BTree:
        """The root B-tree for the filesystem."""
        if self.is_encrypted and self._cipher is None:
            raise Error("Volume is encrypted, unlock volume first")

        return BTree.from_address(
            self.container,
            self.omap.lookup(self.object.apfs_root_tree_oid, self.xid),
            cipher=self._cipher,
        )

    @cached_property
    def snap_meta_tree(self) -> BTree:
        """The snapshot metadata B-tree for the filesystem."""
        return BTree.from_address(
            self.container,
            self.object.apfs_snap_meta_tree_oid,
        )

    @cached_property
    def uuid(self) -> UUID:
        """The filesystem UUID."""
        return UUID(bytes_le=self.object.apfs_vol_uuid)

    @cached_property
    def mtime(self) -> datetime:
        """The last modification time of the filesystem."""
        return from_unix(self.object.apfs_last_mod_time)

    @cached_property
    def flags(self) -> c_apfs.APFS_FS:
        return c_apfs.APFS_FS(self.object.apfs_fs_flags)

    @cached_property
    def is_unencrypted(self) -> bool:
        """Whether the filesystem is unencrypted."""
        return c_apfs.APFS_FS.UNENCRYPTED in self.flags

    @cached_property
    def is_encrypted(self) -> bool:
        """Whether the filesystem is encrypted."""
        return (
            not self.is_unencrypted
            # Related to FileVault, but does not imply full volume encryption
            and c_apfs.APFS_FS.PFK not in self.flags
        )

    @cached_property
    def is_onekey(self) -> bool:
        """Whether the filesystem uses the volume encryption key for all."""
        return c_apfs.APFS_FS.ONEKEY in self.flags

    @cached_property
    def formatted_by(self) -> tuple[str, datetime, int]:
        """Information about the tool that formatted the filesystem."""
        return (
            self.object.apfs_formatted_by.id.split(b"\x00", 1)[0].decode(),
            from_unix(self.object.apfs_formatted_by.timestamp),
            self.object.apfs_formatted_by.last_xid,
        )

    @cached_property
    def modified_by(self) -> list[tuple[str, datetime, int]]:
        """Information about the tools that modified the filesystem."""
        result = []
        for entry in self.object.apfs_modified_by:
            if entry.id == b"\x00" * 64:
                continue

            result.append(
                (
                    entry.id.split(b"\x00", 1)[0].decode(),
                    from_unix(entry.timestamp),
                    entry.last_xid,
                )
            )

        return result

    @cached_property
    def name(self) -> str:
        """The volume name."""
        return self.object.apfs_volname.split(b"\x00", 1)[0].decode()

    @cached_property
    def role(self) -> c_apfs.APFS_VOL_ROLE:
        """The volume role."""
        return c_apfs.APFS_VOL_ROLE(self.object.apfs_role)

    @cached_property
    def snapshots(self) -> list[Snapshot]:
        """All snapshots in the filesystem."""
        result = []

        cursor = Cursor(self.snap_meta_tree)
        for key, value in cursor.walk():
            xid, type = parse_fs_object_key(key)
            if type != c_apfs.APFS_TYPE.SNAP_METADATA:
                continue
            result.append(Snapshot(self, xid, value))

        return result

    @cached_property
    def fext_tree(self) -> BTree:
        """The file extent B-tree for the filesystem."""
        return BTree.from_address(
            self.container,
            self.object.apfs_fext_tree_oid,
        )

    @cached_property
    def keybag(self) -> VolumeKeybag | None:
        """The volume keybag, if present."""
        return self.container.keybag.volume_keybag(self.object.apfs_vol_uuid)

    @cached_property
    def password_hint(self) -> str | None:
        """The password hint for the volume, if present."""
        if not self.keybag:
            return None

        return self.keybag.password_hint(self.object.apfs_vol_uuid)

    def unlock(self, password: str, uuid: UUID | str | None = None) -> None:
        if (vek := self.container.keybag.vek(self.object.apfs_vol_uuid)) is None:
            raise Error("No VEK found for this volume")

        for kek in self.keybag.keks():
            if uuid is not None and str(kek.uuid) != str(uuid):
                continue

            if not kek.verify():
                continue

            try:
                vek = vek.unwrap(kek.unwrap(password))
            except Exception:
                continue
            else:
                self._cipher = create_cipher("aes-xts-128", vek)
                break

        if self._cipher is None:
            raise Error("Failed to unlock volume with the given password")

    def cursor(self) -> Cursor:
        """Create a new cursor for the volume's root B-tree."""
        return Cursor(self.root_tree, self.omap, self.object.apfs_root_tree_oid if self.is_sealed else 0, self.xid)

    def _cursor_state(self, oid: int) -> tuple[BTreeNode, int, list[tuple[BTreeNode, int]]]:
        """Precompute the cursor state for a given object ID.

        Args:
            oid: The object ID to position the cursor at.
        """
        cursor = self.cursor()
        cursor.search((oid, c_apfs.APFS_TYPE.ANY.value), cmp=cmp_fs)

        return cursor.state

    def _cursor(self, oid: int) -> Cursor:
        """Get a cursor positioned at the first record for a given object ID.

        Args:
            oid: The object ID to position the cursor at.
        """
        cursor = self.cursor()
        cursor.state = self._cursor_state(oid)
        return cursor

    def _records(self, oid: int, type: c_apfs.APFS_TYPE) -> Iterator[tuple[c_apfs.APFS_TYPE, bytes, bytes]]:
        """Iterate over all records for a given object ID and type.

        Args:
            oid: The object ID to search for.
            type: The object type to search for, or ``APFS_TYPE.ANY`` for all types.
        """
        cursor = self._cursor(oid)

        while True:
            current_oid, current_type = parse_fs_object_key(cursor.key())
            if current_oid < oid:
                if not cursor.next():
                    break
                continue

            if current_oid > oid or (type != c_apfs.APFS_TYPE.ANY and current_type > type):
                break

            if type == c_apfs.APFS_TYPE.ANY or current_type == type:
                yield current_type, cursor.key(), cursor.value()

            if not cursor.next():
                break

    def records(self, oid: int) -> dict[c_apfs.APFS_TYPE, list[Any]]:
        """Get all records for a given object ID.

        Args:
            oid: The object ID to search for.
        """
        result = {}
        for type, key, value in self._records(oid, c_apfs.APFS_TYPE.ANY):
            key_struct, value_struct = FILESYSTEM_OBJECT_TYPE_MAP[type]
            result.setdefault(type, []).append((key_struct(key), value_struct(value)))
        return result

    def inode(self, oid: int | str, sibling_id: int | None = None) -> INode:
        """Get an inode by its object ID.

        Args:
            oid: The object ID of the inode to retrieve.
            sibling_id: The sibling ID of the inode to retrieve, if applicable.
        """
        if isinstance(oid, str):
            if ":" not in oid:
                raise ValueError("Invalid inode string, expected format 'oid:sibling_id'")
            oid, sibling_id = map(int, oid.split(":", 1))
        return INode(self, oid, sibling_id)

    def inodes(self) -> Iterator[INode]:
        """Iterate over all inodes in the filesystem."""
        cursor = self.cursor()

        for key, value in cursor.walk():
            oid, type = parse_fs_object_key(key)
            if type == c_apfs.APFS_TYPE.INODE:
                inode = self.inode(oid)
                inode._inode_raw = value
                yield inode

    def get(self, path: str | int | DirectoryEntry, node: INode | None = None) -> INode:
        """Get an inode by its path, object ID, or directory entry.

        Args:
            path: The path, object ID, or directory entry of the inode to retrieve.
            node: The starting inode for relative paths. Defaults to the root inode.
        """
        if isinstance(path, int):
            return self.inode(path)

        if isinstance(path, DirectoryEntry):
            return self.inode(path.file_id)

        node = node or self.root
        for p in path.split("/"):
            if not p:
                continue

            if p == ".":
                continue

            if p == "..":
                node = node.parent
                continue

            try:
                node = node.get(p).inode
            except FileNotFoundError:
                raise FileNotFoundError(f"File not found: {path}")

        return node


class Snapshot:
    def __init__(self, fs: FS, xid: int, value: bytes):
        self.fs = fs
        self.xid = xid
        self.value = c_apfs.j_snap_metadata_val(value)

    def __repr__(self) -> str:
        return (
            f"<Snapshot xid={self.xid} name={self.name!r} "
            f"create_time={self.create_time} change_time={self.change_time}>"
        )

    @cached_property
    def create_time(self) -> datetime:
        """The creation time of the snapshot."""
        return from_unix_ns(self.value.create_time)

    @cached_property
    def change_time(self) -> datetime:
        """The change time of the snapshot."""
        return from_unix_ns(self.value.change_time)

    @cached_property
    def name(self) -> str:
        """The name of the snapshot."""
        return self.value.name.decode().rstrip("\x00")

    def open(self) -> FS:
        """Open the snapshot as a filesystem."""
        fs = FS.from_address(
            self.fs.container,
            self.value.sblock_oid,
        )
        fs.omap = self.fs.omap
        return fs


class INode:
    """APFS inode.

    Represents a file or directory in an APFS volume.

    Args:
        volume: Parent APFS volume.
        oid: The object ID of the inode.
        sibling_id: The sibling ID of the inode, if applicable.
    """

    def __init__(self, volume: FS, oid: int, sibling_id: int | None = None):
        self.volume = volume
        self.oid = oid
        self.sibling_id = sibling_id

    def __repr__(self) -> str:
        oid = f"{self.oid}:{self.sibling_id}" if self.sibling_id is not None else str(self.oid)
        return f"<INode oid={oid} name={self.name!r}>"

    @cached_property
    def records(self) -> dict[c_apfs.APFS_TYPE, list[Any]]:
        """All records for this inode."""
        return self.volume.records(self.oid)

    @cached_property
    def _inode_raw(self) -> bytes:
        """The raw inode data."""
        _, _, value = next(self.volume._records(self.oid, c_apfs.APFS_TYPE.INODE), (None, None, None))
        if value is None:
            raise KeyError(f"INode {self.oid} not found in volume")
        return value

    @cached_property
    def inode(self) -> c_apfs.j_inode_val:
        """The parsed inode structure."""
        return c_apfs.j_inode_val(self._inode_raw)

    @cached_property
    def xfields(self) -> dict:
        """The extended fields for this inode."""
        if len(self._inode_raw) == len(c_apfs.j_inode_val):
            return {}

        result = {}
        for field, value in _read_xfields(self._inode_raw[len(c_apfs.j_inode_val) :]):
            type = c_apfs.INO_EXT_TYPE(field.x_type)
            result[type] = (c_apfs.XF(field.x_flags), XF_MAP.get(type, lambda buf: buf)(value))

        return result

    @cached_property
    def xattr(self) -> dict[str, XAttr]:
        """The extended attributes for this inode."""
        result = {}

        for _, key, value in self.volume._records(self.oid, c_apfs.APFS_TYPE.XATTR):
            xattr = XAttr(self, key, value)
            result[xattr.name] = xattr

        return result

    @property
    def parent(self) -> INode:
        """The parent inode."""
        if self.oid in (c_apfs.ROOT_DIR_INO_NUM, c_apfs.PRIV_DIR_INO_NUM):
            return self

        if self.sibling_id is not None:
            parent_id, _ = self.sibling_link or (None, None)
        else:
            parent_id = self.inode.parent_id

        return self.volume.inode(parent_id)

    @property
    def parents(self) -> Iterator[INode]:
        """Iterate over the parent inodes of this inode, up to the root."""
        obj = self
        while obj.parent is not obj:
            obj = obj.parent
            yield obj

    @cached_property
    def private_id(self) -> int:
        """The private ID of this inode."""
        return self.inode.private_id

    @cached_property
    def btime(self) -> datetime:
        """The birth time of this inode."""
        return from_unix_ns(self.inode.create_time)

    @cached_property
    def mtime(self) -> datetime:
        """The modification time of this inode."""
        return from_unix_ns(self.inode.mod_time)

    @cached_property
    def ctime(self) -> datetime:
        """The change time of this inode."""
        return from_unix_ns(self.inode.change_time)

    @cached_property
    def atime(self) -> datetime:
        """The access time of this inode."""
        return from_unix_ns(self.inode.access_time)

    @cached_property
    def internal_flags(self) -> c_apfs.INODE:
        """The inode internal flags."""
        return c_apfs.INODE(self.inode.internal_flags)

    @cached_property
    def bsd_flags(self) -> int:
        """The inode BSD flags."""
        return self.inode.bsd_flags

    def is_compressed(self) -> bool:
        """Return whether this inode is compressed."""
        return bool(self.bsd_flags & c_apfs.UF_COMPRESSED)

    @cached_property
    def uid(self) -> int:
        """The owner user ID of this inode."""
        return self.inode.owner

    @cached_property
    def gid(self) -> int:
        """The owner group ID of this inode."""
        return self.inode.group

    @cached_property
    def mode(self) -> int:
        """The file mode of this inode."""
        return self.inode.mode

    @cached_property
    def type(self) -> int:
        """The file type of this inode."""
        return stat.S_IFMT(self.mode)

    def is_dir(self) -> bool:
        """Return whether this inode is a directory."""
        return stat.S_ISDIR(self.mode)

    def is_file(self) -> bool:
        """Return whether this inode is a regular file."""
        return stat.S_ISREG(self.mode)

    def is_symlink(self) -> bool:
        """Return whether this inode is a symbolic link."""
        return stat.S_ISLNK(self.mode)

    def is_block_device(self) -> bool:
        """Return whether this inode is a block device."""
        return stat.S_ISBLK(self.mode)

    def is_character_device(self) -> bool:
        """Return whether this inode is a character device."""
        return stat.S_ISCHR(self.mode)

    def is_device(self) -> bool:
        """Return whether this inode is a device (block or character)."""
        return self.is_block_device() or self.is_character_device()

    def is_fifo(self) -> bool:
        """Return whether this inode is a FIFO."""
        return stat.S_ISFIFO(self.mode)

    def is_socket(self) -> bool:
        """Return whether this inode is a socket."""
        return stat.S_ISSOCK(self.mode)

    def is_whiteout(self) -> bool:
        """Return whether this inode is a whiteout."""
        return stat.S_ISWHT(self.mode)

    @cached_property
    def size(self) -> int:
        """The size of this inode in bytes."""
        if c_apfs.INODE.HAS_UNCOMPRESSED_SIZE in self.internal_flags:
            return self.inode.uncompressed_size

        if decmpfs := self.xattr.get("com.apple.decmpfs"):
            header = c_apfs.decmpfs_header(decmpfs.open())
            return header.uncompressed_size

        if dstream := self.xfields.get(c_apfs.INO_EXT_TYPE.DSTREAM):
            _, dstream = dstream
            return dstream.size

        return 0

    @cached_property
    def siblings(self) -> list[INode]:
        """All sibling inodes of this inode."""
        result = []
        for _, key, _ in self.volume._records(self.oid, c_apfs.APFS_TYPE.SIBLING_LINK):
            key = c_apfs.j_sibling_key(key)
            if self.sibling_id is not None and key.sibling_id == self.sibling_id:
                continue
            result.append(self.volume.inode(self.oid, key.sibling_id))
        return result

    @cached_property
    def sibling_link(self) -> tuple[int, str] | None:
        """The sibling link (``parent_id``, ``name``) tuple of this inode, if available."""
        if self.sibling_id is not None:
            for _, key, value in self.volume._records(self.oid, c_apfs.APFS_TYPE.SIBLING_LINK):
                if c_apfs.j_sibling_key(key).sibling_id == self.sibling_id:
                    value = c_apfs.j_sibling_val(value)
                    return value.parent_id, value.name.decode().rstrip("\x00")
        return None

    @cached_property
    def name(self) -> str | None:
        """The name of this inode, if available."""
        name = None
        if self.sibling_id is not None:
            _, name = self.sibling_link or (None, None)
        else:
            if c_apfs.INO_EXT_TYPE.NAME in self.xfields:
                _, name = self.xfields[c_apfs.INO_EXT_TYPE.NAME]
                name = name.decode().rstrip("\x00")
        return name

    @cached_property
    def names(self) -> list[str]:
        """All names of this inode, if available."""
        names = {self.name} | {sibling.name for sibling in self.siblings if sibling.name}
        return list(names)

    @cached_property
    def path(self) -> str:
        """The full path of this inode, if available."""
        parts = [self.name or f"<unlinked:{self.oid}>"]
        parts.extend(parent.name or f"<unlinked:{parent.oid}>" for parent in self.parents)

        return "/" + "/".join(parts[::-1])

    @cached_property
    def paths(self) -> list[str]:
        """All full paths of this inode, if available."""
        return [self.path] + [sibling.path for sibling in self.siblings]

    def get(self, name: str) -> DirectoryEntry:
        """Get a directory entry by name."""
        if not self.volume.is_case_insensitive and not self.volume.is_normalization_insensitive:
            # APFS beta's didn't have normalization insensitivity, so when the filesystem is case sensitive
            # we can do a simple exact match
            key = ((self.oid, c_apfs.APFS_TYPE.DIR_REC.value), name.encode() + b"\x00")
            cmp = cmp_fs_dir
        else:
            # Length is not used in the key comparison
            name_hash = (
                _hash_filename(name, self.volume.is_case_insensitive) << c_apfs.J_DREC_HASH_SHIFT
            ) & c_apfs.J_DREC_HASH_MASK
            # If the volume is case sensitive, we can use the name in the search key for an exact match
            # Otherwise, we set it to None to ignore it in the comparison
            name_search = None if self.volume.is_case_insensitive else (name.encode() + b"\x00")

            key = ((self.oid, c_apfs.APFS_TYPE.DIR_REC.value), name_hash, name_search)
            cmp = cmp_fs_dir_hash

        cursor = self.volume.cursor()
        try:
            cursor.search(key, exact=True, cmp=cmp)
        except KeyError:
            raise FileNotFoundError(f"File not found: {name}")

        if not self.volume.is_case_insensitive:
            # On case sensitive volumes, the search is already exact, so no need to check further
            return DirectoryEntry(self.volume, cursor.key(), cursor.value())

        lname = name.casefold()
        while True:
            try:
                dirent = DirectoryEntry(self.volume, cursor.key(), cursor.value())
            except Exception:
                raise FileNotFoundError(f"File not found: {name}")

            # To deal with the possibility of hash collisions, verify the name matches
            oid = dirent.key.hdr.obj_id_and_type & c_apfs.OBJ_ID_MASK
            type = (dirent.key.hdr.obj_id_and_type & c_apfs.OBJ_TYPE_MASK) >> c_apfs.OBJ_TYPE_SHIFT
            if oid != self.oid or type != c_apfs.APFS_TYPE.DIR_REC:
                raise FileNotFoundError(f"File not found: {name}")

            if lname == dirent.name.casefold():
                return dirent

            if not cursor.next():
                raise FileNotFoundError(f"File not found: {name}")

    def listdir(self) -> dict[str, DirectoryEntry]:
        """List the directory entries in this inode."""
        return {e.name: e for e in self.iterdir()}

    def iterdir(self) -> Iterator[DirectoryEntry]:
        """Iterate over the directory entries in this inode."""
        if not self.is_dir():
            raise NotADirectoryError(f"{self} is not a directory")

        for _, key, value in self.volume._records(self.oid, c_apfs.APFS_TYPE.DIR_REC):
            yield DirectoryEntry(self.volume, key, value)

    def readlink(self) -> str:
        """The target of this inode if it is a symbolic link."""
        if not self.is_symlink():
            raise NotASymlinkError(f"{self} is not a symlink")

        if c_apfs.SYMLINK_EA_NAME not in self.xattr:
            raise Error(f"Symlink {self} does not have SYMLINK_EA_NAME xattr")

        return self.xattr[c_apfs.SYMLINK_EA_NAME].open().read().decode().rstrip("\x00")

    def open(self) -> BufferedStream | DecmpfsStream | FileStream:
        """Open a stream for reading the inode data."""
        if self.is_compressed():
            return DecmpfsStream(self)

        if dstream := self.xfields.get(c_apfs.INO_EXT_TYPE.DSTREAM):
            _, dstream = dstream
            return FileStream(self.volume, self.oid, dstream.size)

        return BufferedStream(BytesIO(b""), size=0)


class DirectoryEntry:
    """APFS directory entry.

    Args:
        volume: Parent APFS volume.
        key: The raw directory entry key.
        value: The raw directory entry value.
    """

    def __init__(self, volume: FS, key: bytes, value: bytes):
        self.volume = volume
        if self.volume.is_case_insensitive or self.volume.is_normalization_insensitive:
            self.key = c_apfs.j_drec_hashed_key(key)
        else:
            self.key = c_apfs.j_drec_key(key)
        self.value = c_apfs.j_drec_val(value)

        self.xfields = {}
        if len(value) != len(c_apfs.j_drec_val):
            for field, data in _read_xfields(value[len(c_apfs.j_drec_val) :]):
                type = c_apfs.DREC_EXT_TYPE(field.x_type)
                self.xfields[type] = (c_apfs.XF(field.x_flags), XF_MAP.get(type, lambda buf: buf)(data))

    def __repr__(self) -> str:
        return f"<DirectoryEntry name={self.name!r} file_id={self.file_id} dt={self.dt.name}>"

    @cached_property
    def name(self) -> str:
        """The name of this directory entry."""
        return self.key.name.decode().rstrip("\x00")

    @cached_property
    def hash(self) -> int | None:
        """The hash of this directory entry, if available."""
        if isinstance(self.key, c_apfs.j_drec_key):
            return None
        return (self.key.name_len_and_hash & c_apfs.J_DREC_HASH_MASK) >> c_apfs.J_DREC_HASH_SHIFT

    @cached_property
    def file_id(self) -> int:
        """The object ID of the inode this directory entry refers to."""
        return self.value.file_id

    @cached_property
    def sibling_id(self) -> int | None:
        """The sibling ID of the inode this directory entry refers to, if available."""
        return self.xfields.get(c_apfs.DREC_EXT_TYPE.SIBLING_ID, (None, None))[1]

    @property
    def inode(self) -> INode:
        """The inode this directory entry refers to."""
        return self.volume.inode(self.file_id, self.sibling_id)

    @cached_property
    def date_added(self) -> datetime:
        """The date and time this directory entry was added."""
        return from_unix_ns(self.value.date_added)

    @cached_property
    def dt(self) -> c_apfs.DT:
        """The directory entry type."""
        return c_apfs.DT(self.value.flags & c_apfs.DREC_TYPE_MASK)

    @cached_property
    def type(self) -> int:
        """The file type of this directory entry."""
        return (self.value.flags & c_apfs.DREC_TYPE_MASK) << 12

    def is_dir(self) -> bool:
        """Return whether this directory entry is a directory."""
        return stat.S_ISDIR(self.type)

    def is_file(self) -> bool:
        """Return whether this directory entry is a regular file."""
        return stat.S_ISREG(self.type)

    def is_symlink(self) -> bool:
        """Return whether this directory entry is a symbolic link."""
        return stat.S_ISLNK(self.type)

    def is_block_device(self) -> bool:
        """Return whether this directory entry is a block device."""
        return stat.S_ISBLK(self.type)

    def is_character_device(self) -> bool:
        """Return whether this directory entry is a character device."""
        return stat.S_ISCHR(self.type)

    def is_device(self) -> bool:
        """Return whether this directory entry is a device (block or character)."""
        return self.is_block_device() or self.is_character_device()

    def is_fifo(self) -> bool:
        """Return whether this directory entry is a FIFO."""
        return stat.S_ISFIFO(self.type)

    def is_socket(self) -> bool:
        """Return whether this directory entry is a socket."""
        return stat.S_ISSOCK(self.type)

    def is_whiteout(self) -> bool:
        """Return whether this directory entry is a whiteout."""
        return stat.S_ISWHT(self.type)


class XAttr:
    """APFS extended attribute.

    Args:
        inode: The inode this xattr belongs to.
        key: The raw xattr key.
        value: The raw xattr value.
    """

    def __init__(self, inode: INode, key: bytes, value: bytes):
        self.inode = inode

        self.key = c_apfs.j_xattr_key(key)
        self.value = c_apfs.j_xattr_val(value)

    def __repr__(self) -> str:
        return f"<XAttr name={self.name!r} flags={self.flags}>"

    @cached_property
    def name(self) -> str:
        """The name of this xattr."""
        return self.key.name.decode().rstrip("\x00")

    @cached_property
    def flags(self) -> c_apfs.XATTR:
        """The flags of this xattr."""
        return c_apfs.XATTR(self.value.flags)

    def open(self) -> BufferedStream | FileStream:
        """Open a stream for reading the xattr data."""
        if c_apfs.XATTR.DATA_STREAM in self.flags:
            dstream = c_apfs.j_xattr_dstream(self.value.xdata)
            return FileStream(self.inode.volume, dstream.xattr_obj_id, dstream.dstream.size)

        return BufferedStream(BytesIO(self.value.xdata), size=self.value.xdata_len)


def _read_xfields(data: bytes) -> Iterator[tuple[c_apfs.x_field, bytes]]:
    """Read extended fields from a buffer."""
    blob = c_apfs.xf_blob(data)
    buf = BytesIO(blob.xf_data)
    for field in blob.xf_exts:
        yield field, buf.read(field.x_size)
        # Align to 8 bytes
        buf.seek((buf.tell() + 7) & (-8))


def _hash_filename(name: str, casefold: bool) -> int:
    """Hash a filename according to APFS rules."""
    normalized = unicodedata.normalize("NFD", name)
    if casefold:
        normalized = normalized.casefold()
    return crc32c(normalized.encode("utf-32-le")) ^ 0xFFFFFFFF



from __future__ import annotations

import struct
import zlib
from typing import TYPE_CHECKING

from dissect.util.compression import lzbitmap, lzfse, lzvn
from dissect.util.stream import AlignedStream

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.cursor import Cursor
from dissect.apfs.exception import Error
from dissect.apfs.util import cmp_fext, cmp_fs_extent

if TYPE_CHECKING:
    from dissect.apfs.objects.fs import FS, INode


class FileStream(AlignedStream):
    """APFS file data stream.

    Args:
        volume: APFS volume.
        oid: The object ID of the file.
        size: The size of the file in bytes.
    """

    def __init__(self, volume: FS, oid: int, size: int):
        self.volume = volume
        self.oid = oid

        if self.volume.is_sealed:
            self._cursor = Cursor(self.volume.fext_tree)
        else:
            self._cursor = self.volume.cursor()

        super().__init__(size, self.volume.container.block_size)

    def _lookup(self, offset: int) -> tuple[int, int, int, int]:
        if self.volume.is_sealed:
            return self._lookup_sealed(offset)

        return self._lookup_normal(offset)

    def _lookup_normal(self, offset: int) -> tuple[int, int, int, int]:
        self._cursor.reset().search(
            (
                (self.oid, c_apfs.APFS_TYPE.FILE_EXTENT.value),
                offset,
            ),
            cmp=cmp_fs_extent,
        )

        key = c_apfs.j_file_extent_key(self._cursor.key())
        oid = key.hdr.obj_id_and_type & c_apfs.OBJ_ID_MASK
        type = (key.hdr.obj_id_and_type & c_apfs.OBJ_TYPE_MASK) >> c_apfs.OBJ_TYPE_SHIFT

        if oid != self.oid or type != c_apfs.APFS_TYPE.FILE_EXTENT or key.logical_addr > offset:
            raise Error(f"Could not find file extent for {self.oid} at offset {offset}")

        value = c_apfs.j_file_extent_val(self._cursor.value())

        return (
            key.logical_addr,
            value.phys_block_num,
            value.len_and_flags & c_apfs.J_FILE_EXTENT_LEN_MASK,
            value.crypto_id,
        )

    def _lookup_sealed(self, offset: int) -> tuple[int, int, int, int]:
        self._cursor.reset().search((self.oid, offset), cmp=cmp_fext)

        key = c_apfs.fext_tree_key(self._cursor.key())
        if key.private_id != self.oid or key.logical_addr > offset:
            raise Error(f"Could not find file extent for {self.oid} at offset {offset}")

        value = c_apfs.fext_tree_val(self._cursor.value())

        return (
            key.logical_addr,
            value.phys_block_num,
            value.len_and_flags & c_apfs.J_FILE_EXTENT_LEN_MASK,
            0,
        )

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        while length:
            logical_address, physical_address, extent_length, crypto_id = self._lookup(offset)
            block = self.volume.container._read_block(physical_address, extent_length // self.align)

            if self.volume.is_encrypted:
                if not self.volume._cipher:
                    raise Error("Volume is encrypted, unlock it first")

                if self.volume.is_onekey:
                    block = self.volume._cipher.decrypt(block, crypto_id * self.volume.container.sectors_per_block)
                else:
                    raise Error("Multi-key encryption is not supported yet")

            if offset_in_extent := offset - logical_address:
                block = block[offset_in_extent:]

            if length < len(block):
                block = block[:length]

            result.append(block)
            offset += min(extent_length, length)
            length -= min(extent_length, length)

        return b"".join(result)


class DecmpfsStream(AlignedStream):
    """APFS decompressed file data stream.

    Args:
        inode: The inode of the compressed file.
    """

    def __init__(self, inode: INode):
        self.inode = inode

        if (attr := inode.xattr.get("com.apple.decmpfs")) is None:
            raise Error(f"{inode} is marked as compressed, but has no decmpfs xattr")

        buf = attr.open()
        self._header = c_apfs.decmpfs_header(buf)
        if self._header.magic.to_bytes(4, "big") != c_apfs.DECMPFS_MAGIC:
            raise Error(
                "Invalid decmpfs magic "
                f"(expected {c_apfs.DECMPFS_MAGIC!r}, got {self._header.magic.to_bytes(4, 'big')!r})"
            )

        self._fork = None
        self._entries = []

        if self._header.algorithm & 1 == 0:
            # In resource fork
            if (fork := inode.xattr.get("com.apple.ResourceFork")) is None:
                raise Error(f"{inode} is missing resource fork for decmpfs")

            self._fork = fork.open()

            if self._header.algorithm in (c_apfs.DECMPFS_ZLIB_RSRC, c_apfs.DECMPFS_PLAIN_RSRC):
                # data_offset, mgmt_offset, data_size, mgmt_size
                self._fork_data_offset, _, _, _ = struct.unpack(">IIII", self._fork.read(16))

                self._fork.seek(self._fork_data_offset)
                for _ in range(struct.unpack("<II", self._fork.read(8))[1]):
                    offset, length = struct.unpack("<II", self._fork.read(8))
                    self._entries.append((self._fork_data_offset + 4 + offset, length))
            else:
                offsets = c_apfs.uint32_t[
                    ((self._header.uncompressed_size + (c_apfs.DECMPFS_BLOCK_SIZE - 1)) // c_apfs.DECMPFS_BLOCK_SIZE)
                    + 1
                ](self._fork)
                for i in range(len(offsets) - 1):
                    self._entries.append((offsets[i], offsets[i + 1] - offsets[i]))
        else:
            # In attribute
            self._data = buf.read()

        super().__init__(self._header.uncompressed_size, align=c_apfs.DECMPFS_BLOCK_SIZE)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        algorithm = self._header.algorithm
        block = offset // self.align
        while length:
            if self._fork is not None:
                entry_offset, entry_length = self._entries[block]

                self._fork.seek(entry_offset)
                chunk = self._fork.read(entry_length)
            else:
                chunk = self._data

            if algorithm in (c_apfs.DECMPFS_ZLIB_ATTR, c_apfs.DECMPFS_ZLIB_RSRC):
                if chunk[0] == 0x78 and len(chunk) >= 2:
                    chunk = zlib.decompress(chunk)
                elif chunk[0] & 0x0F == 0x0F:
                    chunk = chunk[1:]
                else:
                    raise Error(f"Invalid zlib chunk: {chunk[0]:#x}")
            elif algorithm in (c_apfs.DECMPFS_LZVN_ATTR, c_apfs.DECMPFS_LZVN_RSRC):
                chunk = chunk[1:] if chunk[0] == 0x06 else lzvn.decompress(chunk)
            elif algorithm in (c_apfs.DECMPFS_PLAIN_ATTR, c_apfs.DECMPFS_PLAIN_RSRC):
                chunk = chunk[1:]
            elif algorithm in (c_apfs.DECMPFS_LZFSE_ATTR, c_apfs.DECMPFS_LZFSE_RSRC):
                chunk = chunk[1:] if chunk[0] == 0xFF else lzfse.decompress(chunk)
            elif algorithm in (c_apfs.DECMPFS_LZBITMAP_ATTR, c_apfs.DECMPFS_LZBITMAP_RSRC):
                chunk = chunk[1:] if chunk[0] & 0x0F == 0x0F else lzbitmap.decompress(chunk)
            else:
                raise Error(f"Unsupported decmpfs algorithm {algorithm}")

            result.append(chunk)

            offset += self.align
            length -= min(self.align, length)
            block += 1

        return b"".join(result)



from __future__ import annotations

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class NxFusionWbc(Object):
    """APFS Fusion Write-Back Cache object."""

    __type__ = c_apfs.OBJECT_TYPE_NX_FUSION_WBC
    __struct__ = c_apfs.fusion_wbc_phys
    object: c_apfs.fusion_wbc_phys



from __future__ import annotations

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class NxReapList(Object):
    """APFS NX Reap List object."""

    __type__ = c_apfs.OBJECT_TYPE_NX_REAP_LIST
    __struct__ = c_apfs.nx_reap_list_phys
    object: c_apfs.nx_reap_list_phys



from __future__ import annotations

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class SpacemanChunkInfoAddressBlock(Object):
    """APFS Spaceman Chunk Info Address Block object."""

    __type__ = c_apfs.OBJECT_TYPE_SPACEMAN_CAB
    __struct__ = c_apfs.cib_addr_block
    object: c_apfs.cib_addr_block



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from dissect.apfs.objects.btree import cmp_default

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from typing_extensions import Self

    from dissect.apfs.objects import BTree, BTreeNode, ObjectMap


class Cursor:
    """Cursor for traversing B-Trees.

    Args:
        btree: B-Tree to traverse.
        omap: Optional object map to resolve child pointers.
        oid: Optional base object ID to resolve child pointers.
        xid: Optional transaction ID to resolve child pointers.
    """

    def __init__(self, btree: BTree, omap: ObjectMap | None = None, oid: int = 0, xid: int | None = None) -> None:
        self.btree = btree
        self.omap = omap
        self.oid = oid
        self.xid = xid

        self._node = btree.root
        self._idx = 0

        self._stack = []

    @property
    def state(self) -> tuple[BTreeNode, int, list[tuple[BTreeNode, int]]]:
        """Get the current cursor state."""
        return self._node, self._idx, self._stack[:]

    @state.setter
    def state(self, value: tuple[BTreeNode, int, list[tuple[BTreeNode, int]]]) -> None:
        """Set the current cursor state."""
        self._node, self._idx, self._stack = value[0], value[1], value[2][:]

    def reset(self) -> Self:
        """Reset the cursor to the root of the B-Tree."""
        self._node = self.btree.root
        self._idx = 0
        self._stack = []

        return self

    def key(self) -> bytes:
        """Get the current key."""
        return self.btree._read_key(self._node, self._idx)

    def value(self) -> bytes:
        """Get the current value."""
        return self.btree._read_value(self._node, self._idx)

    def item(self) -> tuple[bytes, bytes]:
        """Get the current key and value."""
        return self.key(), self.value()

    def first(self) -> bool:
        """Move the cursor to the first item in the B-Tree."""
        self.reset()
        while self._node.is_nonleaf:
            self.push()

        return self._node.nkeys != 0

    def last(self) -> bool:
        """Move the cursor to the last item in the B-Tree."""
        self.reset()
        while self._node.is_nonleaf:
            self._idx = self._node.nkeys - 1
            self.push()

        self._idx = self._node.nkeys - 1
        return self._node.nkeys != 0

    def next(self) -> bool:
        """Move the cursor to the next item in the B-Tree."""
        if self._node.is_nonleaf:
            # Treat as if we were at the first key
            self.first()
            return self._node.nkeys != 0

        if self._idx + 1 < self._node.nkeys:
            self._idx += 1
        elif self._stack:
            # End of current node, traverse to the next leaf node

            # First pop until we find a node with unvisited keys
            while self._idx + 1 >= self._node.nkeys:
                if not self._stack:
                    return False
                self.pop()

            self._idx += 1

            # Then push down to the next leaf
            while self._node.is_nonleaf:
                self.push()
        else:
            return False

        return True

    def prev(self) -> bool:
        """Move the cursor to the previous item in the B-Tree."""
        if self._node.is_nonleaf:
            # Treat as if we were at the last key
            self.last()
            return self._node.nkeys != 0

        if self._idx - 1 >= 0:
            self._idx -= 1
        elif self._stack:
            # Start of current node, traverse to the previous leaf node

            # First pop until we find a node with unvisited keys
            while self._idx - 1 < 0:
                if not self._stack:
                    # Start of B-Tree reached
                    return False
                self.pop()

            self._idx -= 1

            # Then push down to the rightmost leaf
            while self._node.is_nonleaf:
                self._idx = self._node.nkeys - 1
                self.push()
        else:
            # Start of B-Tree reached
            return False

        return True

    def push(self) -> Self:
        """Push down to the child node at the current index."""
        child_node = self.btree._node_child(self._node, self._idx, self.omap, self.oid, self.xid)

        self._stack.append((self._node, self._idx))
        self._node = child_node
        self._idx = 0

        return self

    def pop(self) -> Self:
        """Pop back to the parent node."""
        if not self._stack:
            raise IndexError("Cannot pop from an empty stack")

        self._node, self._idx = self._stack.pop()

        return self

    def walk(self) -> Iterator[tuple[bytes, bytes]]:
        """Walk the B-Tree in order, yielding (key, value) tuples."""
        if self.first():
            yield self.item()

            while self.next():
                yield self.item()

    def search(
        self,
        key: Any,
        *,
        exact: bool = False,
        cmp: Callable[[Any, bytes], Literal[-1, 0, 1]] = cmp_default,
    ) -> Self:
        """Search for a key in the B-Tree.

        Args:
            key: Key to search for.
            exact: If ``True``, only return if an exact match is found.
            cmp: Comparison function to use. Should return -1, 0, or 1.
        """
        while self._node.is_nonleaf:
            self._idx = self.btree._node_search(self._node, key, cmp=cmp)
            self.push()

        self._idx = self.btree._node_search(self._node, key, exact=exact, cmp=cmp)
        if self._idx >= self._node.nkeys or self._idx == -1:
            raise KeyError(f"Key not found: {key!r}")

        return self



from dissect.apfs.apfs import APFS
from dissect.apfs.exception import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)

__all__ = [
    "APFS",
    "Error",
    "FileNotFoundError",
    "NotADirectoryError",
    "NotASymlinkError",
]



from __future__ import annotations

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class EfiJumpstart(Object):
    """APFS EFI Jumpstart object."""

    __type__ = c_apfs.OBJECT_TYPE_EFI_JUMPSTART
    __struct__ = c_apfs.nx_efi_jumpstart
    object: c_apfs.nx_efi_jumpstart



from __future__ import annotations

import struct
from typing import Literal

from dissect.util.xmemoryview import xmemoryview

from dissect.apfs.c_apfs import c_apfs


def fletcher64(data: bytes) -> int:
    """Compute the Fletcher-64 checksum of the given data."""
    sum1 = 0
    sum2 = 0

    for word in xmemoryview(data, "<I"):
        sum1 = (sum1 + word) & 0xFFFFFFFFFFFFFFFF
        sum2 = (sum2 + sum1) & 0xFFFFFFFFFFFFFFFF

    ck_low = 0xFFFFFFFF - ((sum1 + sum2) % 0xFFFFFFFF)
    ck_high = 0xFFFFFFFF - ((sum1 + ck_low) % 0xFFFFFFFF)

    return ck_low | (ck_high << 32)


def make_fs_object_key(oid: int, type: c_apfs.APFS_TYPE) -> c_apfs.j_key:
    """Create a ``j_key`` struct for the given object ID and type."""
    return c_apfs.j_key((oid & c_apfs.OBJ_ID_MASK) | (type << c_apfs.OBJ_TYPE_SHIFT))


def parse_fs_object_key(data: bytes) -> tuple[int, c_apfs.APFS_TYPE]:
    """Parse an object ID and type from a ``j_key`` struct."""
    key = c_apfs.j_key(data)
    oid = key.obj_id_and_type & c_apfs.OBJ_ID_MASK
    type = (key.obj_id_and_type & c_apfs.OBJ_TYPE_MASK) >> c_apfs.OBJ_TYPE_SHIFT
    return oid, c_apfs.APFS_TYPE(type)


def cmp_default(key: bytes, other: bytes) -> Literal[-1, 0, 1]:
    """Default comparison function for B-tree keys."""
    return (key < other) - (key > other)


_QQ = struct.Struct("<QQ")


def cmp_omap(key: tuple[int, int], other: bytes) -> Literal[-1, 0, 1]:
    """Comparison function for OMAP keys."""
    oid, xid = key
    (other_oid, other_xid) = _QQ.unpack(other)
    if (res := (oid < other_oid) - (oid > other_oid)) != 0:
        return res
    return (xid < other_xid) - (xid > other_xid)


_Q = struct.Struct("<Q")
_OBJ_ID_MASK = c_apfs.OBJ_ID_MASK
_OBJ_TYPE_MASK = c_apfs.OBJ_TYPE_MASK
_OBJ_TYPE_SHIFT = c_apfs.OBJ_TYPE_SHIFT


def cmp_fs(key: tuple[int, int], other: bytes) -> Literal[-1, 0, 1]:
    """Comparison function for FS object keys."""
    obj_id, type = key

    (other_obj_id_and_type,) = _Q.unpack_from(other)
    other_obj_id = other_obj_id_and_type & _OBJ_ID_MASK
    other_type = (other_obj_id_and_type & _OBJ_TYPE_MASK) >> _OBJ_TYPE_SHIFT

    if (res := (obj_id < other_obj_id) - (obj_id > other_obj_id)) != 0:
        return res

    if (res := (type < other_type) - (type > other_type)) != 0:
        return res

    return 0


def cmp_fs_extent(key: tuple[tuple[int, int], int], other: bytes) -> Literal[-1, 0, 1]:
    """Comparison function for FS extent keys."""
    obj_id_and_type, logical_addr = key

    # First compare the j_key portion
    if (res := cmp_fs(obj_id_and_type, other[:8])) != 0:
        return res

    (other_logical_addr,) = _Q.unpack_from(other, 8)

    return (logical_addr < other_logical_addr) - (logical_addr > other_logical_addr)


def cmp_fext(key: tuple[int, int], other: bytes) -> Literal[-1, 0, 1]:
    """Comparison function for sealed file extent keys."""
    private_id, logical_addr = key

    other_private_id, other_logical_addr = _QQ.unpack(other)

    if (res := (private_id < other_private_id) - (private_id > other_private_id)) != 0:
        return res

    return (logical_addr < other_logical_addr) - (logical_addr > other_logical_addr)


_H = struct.Struct("<H")


def cmp_fs_dir(key: tuple[tuple[int, int], bytes], other: bytes) -> Literal[-1, 0, 1]:
    """Comparison function for FS directory entries."""
    # Slightly more unreadable but faster than parsing a struct
    obj_id_and_type, name = key

    # First compare the j_key portion
    if (res := cmp_fs(obj_id_and_type, other[:8])) != 0:
        return res

    # Then compare the name
    (other_name_len,) = _H.unpack_from(other, 8)
    other_name = other[10 : 10 + (other_name_len)]
    return (name < other_name) - (name > other_name)


_I = struct.Struct("<I")
_J_DREC_LEN_MASK = c_apfs.J_DREC_LEN_MASK


def cmp_fs_dir_hash(key: tuple[tuple[int, int], int, bytes | None], other: bytes) -> Literal[-1, 0, 1]:
    """Comparison function for FS directory entries."""
    # Slightly more unreadable but faster than parsing a struct
    obj_id_and_type, name_hash, name = key

    # First compare the j_key portion
    if (res := cmp_fs(obj_id_and_type, other[:8])) != 0:
        return res

    # Then compare the name_len_hash
    (other_name_len_hash,) = _I.unpack_from(other, 8)
    other_hash = other_name_len_hash & c_apfs.J_DREC_HASH_MASK

    if (res := (name_hash < other_hash) - (name_hash > other_hash)) != 0:
        return res

    # Special case for searching without a name
    if name is None:
        return 0

    # Finally compare the name
    other_name = other[12 : 12 + (other_name_len_hash & _J_DREC_LEN_MASK)]
    return (name < other_name) - (name > other_name)



from __future__ import annotations

from dissect.apfs.objects.base import Object
from dissect.apfs.objects.btree import BTree
from dissect.apfs.objects.btree_node import BTreeNode
from dissect.apfs.objects.checkpoint_map import CheckpointMap
from dissect.apfs.objects.efi_jumpstart import EfiJumpstart
from dissect.apfs.objects.er_recovery_block import EncryptionRollingRecoveryBlock
from dissect.apfs.objects.er_state import EncryptionRollingState
from dissect.apfs.objects.fs import FS
from dissect.apfs.objects.gbitmap import GBitmap
from dissect.apfs.objects.gbitmap_block import GBitmapBlock
from dissect.apfs.objects.integrity_meta import IntegrityMeta
from dissect.apfs.objects.nx_fusion_wbc import NxFusionWbc
from dissect.apfs.objects.nx_fusion_wbc_list import NxFusionWbcList
from dissect.apfs.objects.nx_reap_list import NxReapList
from dissect.apfs.objects.nx_reaper import NxReaper
from dissect.apfs.objects.nx_superblock import NxSuperblock
from dissect.apfs.objects.omap import ObjectMap
from dissect.apfs.objects.snap_meta_ext import SnapMetaExt
from dissect.apfs.objects.spaceman import Spaceman
from dissect.apfs.objects.spaceman_bitmap import SpacemanBitmap
from dissect.apfs.objects.spaceman_cab import SpacemanChunkInfoAddressBlock
from dissect.apfs.objects.spaceman_cib import SpacemanChunkInfoBlock

__all__ = [
    "FS",
    "BTree",
    "BTreeNode",
    "CheckpointMap",
    "EfiJumpstart",
    "EncryptionRollingRecoveryBlock",
    "EncryptionRollingState",
    "GBitmap",
    "GBitmapBlock",
    "IntegrityMeta",
    "NxFusionWbc",
    "NxFusionWbcList",
    "NxReapList",
    "NxReaper",
    "NxSuperblock",
    "Object",
    "ObjectMap",
    "SnapMetaExt",
    "Spaceman",
    "SpacemanBitmap",
    "SpacemanChunkInfoAddressBlock",
    "SpacemanChunkInfoBlock",
]



from __future__ import annotations

from functools import cached_property, lru_cache

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object
from dissect.apfs.objects.btree import BTree
from dissect.apfs.util import cmp_omap


class ObjectMap(Object):
    """APFS Object Map (OMAP) object."""

    __type__ = c_apfs.OBJECT_TYPE_OMAP
    __struct__ = c_apfs.omap_phys
    object: c_apfs.omap_phys

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.lookup = lru_cache(128)(self.lookup)

    def is_valid(self) -> bool:
        return (
            super().is_valid()
            and self.type == c_apfs.OBJECT_TYPE.OMAP
            and self.subtype == 0
            and self.is_physical
            and self.oid == self.address
        )

    @cached_property
    def btree(self) -> BTree:
        """The B-tree of the object map."""
        return BTree(self.container, self.object.om_tree_oid)

    def lookup(self, oid: int, xid: int | None = None) -> int:
        """Lookup the physical address of an object by its OID and optional transaction ID.

        Args:
            oid: The object ID to look up.
            xid: Optional transaction ID to look up. If not provided, the first version is returned.
        """
        value = self.btree.search((oid, xid or 0), cmp=cmp_omap)

        omap_val = c_apfs.omap_val(value)
        return omap_val.ov_paddr



from __future__ import annotations

from functools import cached_property

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class BTreeNode(Object):
    """APFS B-tree Node object."""

    __type__ = c_apfs.OBJECT_TYPE_BTREE_NODE
    __struct__ = c_apfs.btree_node_phys
    object: c_apfs.btree_node_phys

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._data_offset = len(self.__struct__)
        self._key_offset = self._data_offset + self.object.btn_table_space.off + self.object.btn_table_space.len

    def __repr__(self) -> str:
        return f"<BTreeNode flags={self.flags} nkeys={self.object.btn_nkeys}>"

    @cached_property
    def toc(self) -> list[c_apfs.kvoff | c_apfs.kvloc]:
        """The B-tree node's table of contents."""
        nkeys = self.object.btn_nkeys
        buf = self.block[self._data_offset + self.object.btn_table_space.off :]

        if self._has_fixed_kv_size:
            return c_apfs.kvoff[nkeys](buf)

        return c_apfs.kvloc[nkeys](buf)

    @cached_property
    def flags(self) -> c_apfs.BTNODE:
        """The B-tree node's flags."""
        return c_apfs.BTNODE(self.object.btn_flags)

    @cached_property
    def is_root(self) -> bool:
        """Whether the node is the root node."""
        return bool(self.object.btn_flags & c_apfs.BTNODE_ROOT)

    @cached_property
    def is_leaf(self) -> bool:
        """Whether the node is a leaf node."""
        return bool(self.object.btn_flags & c_apfs.BTNODE_LEAF)

    @cached_property
    def is_nonleaf(self) -> bool:
        """Whether the node is a non-leaf node."""
        return not self.is_leaf

    @cached_property
    def _has_fixed_kv_size(self) -> bool:
        """Whether the B-tree node has fixed key/value sizes."""
        return bool(self.object.btn_flags & c_apfs.BTNODE_FIXED_KV_SIZE)

    @cached_property
    def level(self) -> int:
        """The B-tree node's level."""
        return self.object.btn_level

    @cached_property
    def nkeys(self) -> int:
        """The number of keys in the B-tree node."""
        return self.object.btn_nkeys



from __future__ import annotations

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class GBitmapBlock(Object):
    """APFS General Bitmap Block object."""

    __type__ = c_apfs.OBJECT_TYPE_GBITMAP_BLOCK
    __struct__ = c_apfs.gbitmap_block_phys
    object: c_apfs.gbitmap_block_phys



from __future__ import annotations

from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, Any, Literal

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object
from dissect.apfs.objects.btree_node import BTreeNode
from dissect.apfs.util import cmp_default

if TYPE_CHECKING:
    from collections.abc import Callable

    from dissect.fve.crypto import Cipher

    from dissect.apfs.apfs import APFS
    from dissect.apfs.objects.omap import ObjectMap


class BTree(Object):
    """APFS B-tree object."""

    __type__ = c_apfs.OBJECT_TYPE_BTREE
    object: c_apfs.btree_node_phys

    def __init__(self, container: APFS, address: int, *, block: bytes | None = None, cipher: Cipher | None = None):
        # Break object model a bit here for convenience of having the root node as a separate object
        self.container = container
        self.address = address

        self.root = BTreeNode(container, address, block=block, cipher=cipher)
        self.block = self.root.block
        self.cipher = self.root.cipher
        self.object = self.root.object
        self.header = self.root.header

        # Root nodes have a btree_info struct at the end of the block
        self.info = c_apfs.btree_info(self.root.block[-len(c_apfs.btree_info) :])

        self._read_key = lru_cache(128)(self._read_key)
        self._node_child = lru_cache(128)(self._node_child)
        self._node_search = lru_cache(128)(self._node_search)

    def __repr__(self) -> str:
        return f"<BTree flags={self.flags}>"

    @cached_property
    def flags(self) -> c_apfs.BTREE:
        """The B-tree's flags."""
        return c_apfs.BTREE(self.info.bt_fixed.bt_flags)

    @cached_property
    def _uses_physical_oid(self) -> bool:
        """Whether the B-tree uses physical OIDs."""
        return bool(self.flags & c_apfs.BTREE_PHYSICAL)

    @cached_property
    def _uses_ephemeral_oid(self) -> bool:
        """Whether the B-tree uses ephemeral OIDs."""
        return bool(self.flags & c_apfs.BTREE_EPHEMERAL)

    @cached_property
    def _uses_virtual_oid(self) -> bool:
        """Whether the B-tree uses virtual OIDs."""
        return not self._uses_physical_oid and not self._uses_ephemeral_oid

    @cached_property
    def _node_size(self) -> int:
        """The B-tree node size."""
        return self.info.bt_fixed.bt_node_size

    @cached_property
    def _key_size(self) -> int:
        """The B-tree key size."""
        return self.info.bt_fixed.bt_key_size

    @cached_property
    def _val_size(self) -> int:
        """The B-tree value size."""
        return self.info.bt_fixed.bt_val_size

    def _read_key(self, node: BTreeNode, idx: int) -> bytes:
        """Read a key from a B-tree node.

        Args:
            node: The B-tree node to read from.
            idx: The index of the key to read.
        """
        entry = node.toc[idx]

        if node._has_fixed_kv_size:
            offset = entry.k
            size = self._key_size
        else:
            offset = entry.k.off
            size = entry.k.len

        return bytes(node.block[node._key_offset + offset : node._key_offset + offset + size])

    def _read_value(self, node: BTreeNode, idx: int) -> bytes:
        """Read a value from a B-tree node.

        Args:
            node: The B-tree node to read from.
            idx: The index of the value to read.
        """
        entry = node.toc[idx]

        if node._has_fixed_kv_size:
            offset = entry.v
            size = self._val_size
        else:
            offset = entry.v.off
            size = entry.v.len

        value_offset = self._node_size - (len(c_apfs.btree_info) if node.is_root else 0)
        return bytes(node.block[value_offset - offset : value_offset - offset + size])

    def _node_child(
        self, node: BTreeNode, idx: int, omap: ObjectMap | None = None, oid: int = 0, xid: int | None = None
    ) -> BTreeNode:
        """Get a child node from a B-tree node.

        Args:
            node: The B-tree node to get the child from.
            idx: The index of the child to get.
            omap: The object map to use for resolving virtual child objects.
            oid: The base object ID to use for resolving virtual child objects.
            xid: The transaction ID to use for resolving virtual child objects.
        """
        if node.is_leaf:
            raise TypeError("Cannot traverse from a leaf node")

        child_oid = c_apfs.oid_t(self._read_value(node, idx))

        if self._uses_physical_oid:
            child_node = BTreeNode.from_address(self.container, child_oid, cipher=self.cipher)
        elif self._uses_ephemeral_oid:
            raise NotImplementedError("Ephemeral OIDs are not yet supported")
        elif self._uses_virtual_oid:
            if not omap:
                raise ValueError("Cannot resolve virtual OID without an Object Map")
            child_node = BTreeNode.from_address(self.container, omap.lookup(oid + child_oid, xid), cipher=self.cipher)

        return child_node

    def _node_search(
        self,
        node: BTreeNode,
        key: Any,
        *,
        exact: bool = False,
        cmp: Callable[[Any, bytes], Literal[-1, 0, 1]] = cmp_default,
    ) -> int:
        """Search for a key in a B-tree node.

        Args:
            node: The B-tree node to search.
            key: The key to search for.
            exact: If ``True``, only return if an exact match is found.
            cmp: Comparison function to use. Should return -1, 0, or 1
        """
        lo, hi = 0, node.nkeys - 1

        while lo <= hi:
            mid = (lo + hi) // 2

            res = cmp(key, self._read_key(node, mid))
            if res == 0:
                break

            if res < 0:
                lo = mid + 1
            else:
                hi = mid - 1

        if exact:
            return mid if res == 0 else -1

        return mid if res <= 0 else mid - 1

    def search(
        self,
        key: bytes,
        *,
        exact: bool = False,
        cmp: Callable[[Any, bytes], Literal[-1, 0, 1]] = cmp_default,
        omap: ObjectMap | None = None,
        oid: int = 0,
        xid: int | None = None,
    ) -> bytes | None:
        """Search for a key in the B-tree.

        Args:
            key: Key to search for.
            exact: If ``True``, only return if an exact match is found.
            cmp: Comparison function to use. Should return -1, 0, or 1.
            omap: The object map to use for resolving virtual child objects.
            oid: The base object ID to use for resolving virtual child objects.
            xid: The transaction ID to use for resolving virtual child objects.
        """
        node = self.root

        while node.is_nonleaf:
            idx = self._node_search(node, key, cmp=cmp)
            node = self._node_child(node, idx, omap, oid, xid)

        idx = self._node_search(node, key, exact=exact, cmp=cmp)
        if idx >= node.nkeys or idx == -1:
            raise KeyError(f"Key not found: {key!r}")

        return self._read_value(node, idx)



from __future__ import annotations

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class CheckpointMap(Object):
    """APFS Checkpoint Map object."""

    __type__ = c_apfs.OBJECT_TYPE_CHECKPOINT_MAP
    __struct__ = c_apfs.checkpoint_map_phys
    object: c_apfs.checkpoint_map_phys



# Resources:
# - https://developer.apple.com/support/downloads/Apple-File-System-Reference.pdf
# - https://github.com/sgan81/apfs-fuse
# - https://github.com/linux-apfs/linux-apfs-rw
from dissect.cstruct import cstruct

apfs_def = """
/*
 * Common types
 */

typedef uint16_t mode_t;
typedef uint32_t uid_t;
typedef uint32_t gid_t;

typedef uint32_t cp_key_class_t;
typedef uint32_t cp_key_os_version_t;
typedef uint16_t cp_key_revision_t;
typedef uint32_t crypto_flags_t;

typedef int64_t paddr_t;

struct prange {
    paddr_t     pr_start_paddr;
    uint64_t    pr_block_count;
};
typedef struct prange prange_t;

typedef uint64_t oid_t;
typedef uint64_t xid_t;

/*
 * Objects
 */

/* Object Identifier Constants */
#define OID_NX_SUPERBLOCK               1
#define OID_INVALID                     0
#define OID_RESERVED_COUNT              1024

/* Object Type Masks */
#define OBJECT_TYPE_MASK                0x0000ffff
#define OBJECT_TYPE_FLAGS_MASK          0xffff0000
#define OBJ_STORAGETYPE_MASK            0xc0000000
#define OBJECT_TYPE_FLAGS_DEFINED_MASK  0xf8000000

/* Object Types */
#define OBJECT_TYPE_NX_SUPERBLOCK       0x00000001
#define OBJECT_TYPE_BTREE               0x00000002
#define OBJECT_TYPE_BTREE_NODE          0x00000003
#define OBJECT_TYPE_SPACEMAN            0x00000005
#define OBJECT_TYPE_SPACEMAN_CAB        0x00000006
#define OBJECT_TYPE_SPACEMAN_CIB        0x00000007
#define OBJECT_TYPE_SPACEMAN_BITMAP     0x00000008
#define OBJECT_TYPE_SPACEMAN_FREE_QUEUE 0x00000009
#define OBJECT_TYPE_EXTENT_LIST_TREE    0x0000000a
#define OBJECT_TYPE_OMAP                0x0000000b
#define OBJECT_TYPE_CHECKPOINT_MAP      0x0000000c
#define OBJECT_TYPE_FS                  0x0000000d
#define OBJECT_TYPE_FSTREE              0x0000000e
#define OBJECT_TYPE_BLOCKREFTREE        0x0000000f
#define OBJECT_TYPE_SNAPMETATREE        0x00000010
#define OBJECT_TYPE_NX_REAPER           0x00000011
#define OBJECT_TYPE_NX_REAP_LIST        0x00000012
#define OBJECT_TYPE_OMAP_SNAPSHOT       0x00000013
#define OBJECT_TYPE_EFI_JUMPSTART       0x00000014
#define OBJECT_TYPE_FUSION_MIDDLE_TREE  0x00000015
#define OBJECT_TYPE_NX_FUSION_WBC       0x00000016
#define OBJECT_TYPE_NX_FUSION_WBC_LIST  0x00000017
#define OBJECT_TYPE_ER_STATE            0x00000018
#define OBJECT_TYPE_GBITMAP             0x00000019
#define OBJECT_TYPE_GBITMAP_TREE        0x0000001a
#define OBJECT_TYPE_GBITMAP_BLOCK       0x0000001b
#define OBJECT_TYPE_ER_RECOVERY_BLOCK   0x0000001c
#define OBJECT_TYPE_SNAP_META_EXT       0x0000001d
#define OBJECT_TYPE_INTEGRITY_META      0x0000001e
#define OBJECT_TYPE_FEXT_TREE           0x0000001f
#define OBJECT_TYPE_RESERVED_20         0x00000020
#define OBJECT_TYPE_INVALID             0x00000000
#define OBJECT_TYPE_TEST                0x000000ff
#define OBJECT_TYPE_CONTAINER_KEYBAG    0x6b657973      // 'keys'
#define OBJECT_TYPE_VOLUME_KEYBAG       0x72656373      // 'recs'
#define OBJECT_TYPE_MEDIA_KEYBAG        0x6d6b6579      // 'mkey'

// As a cstruct enum
enum OBJECT_TYPE {
    NX_SUPERBLOCK                       = 0x00000001,
    BTREE                               = 0x00000002,
    BTREE_NODE                          = 0x00000003,
    SPACEMAN                            = 0x00000005,
    SPACEMAN_CAB                        = 0x00000006,
    SPACEMAN_CIB                        = 0x00000007,
    SPACEMAN_BITMAP                     = 0x00000008,
    SPACEMAN_FREE_QUEUE                 = 0x00000009,
    EXTENT_LIST_TREE                    = 0x0000000a,
    OMAP                                = 0x0000000b,
    CHECKPOINT_MAP                      = 0x0000000c,
    FS                                  = 0x0000000d,
    FSTREE                              = 0x0000000e,
    BLOCKREFTREE                        = 0x0000000f,
    SNAPMETATREE                        = 0x00000010,
    NX_REAPER                           = 0x00000011,
    NX_REAP_LIST                        = 0x00000012,
    OMAP_SNAPSHOT                       = 0x00000013,
    EFI_JUMPSTART                       = 0x00000014,
    FUSION_MIDDLE_TREE                  = 0x00000015,
    NX_FUSION_WBC                       = 0x00000016,
    NX_FUSION_WBC_LIST                  = 0x00000017,
    ER_STATE                            = 0x00000018,
    GBITMAP                             = 0x00000019,
    GBITMAP_TREE                        = 0x0000001a,
    GBITMAP_BLOCK                       = 0x0000001b,
    ER_RECOVERY_BLOCK                   = 0x0000001c,
    SNAP_META_EXT                       = 0x0000001d,
    INTEGRITY_META                      = 0x0000001e,
    FEXT_TREE                           = 0x0000001f,
    RESERVED_20                         = 0x00000020,
    INVALID                             = 0x00000000,
    TEST                                = 0x000000ff,
    CONTAINER_KEYBAG                    = 0x6b657973,   // 'keys'
    VOLUME_KEYBAG                       = 0x72656373,   // 'recs'
    MEDIA_KEYBAG                        = 0x6d6b6579,   // 'mkey'
};

/* Object type flags */
#define OBJ_VIRTUAL                     0x00000000
#define OBJ_EPHEMERAL                   0x80000000
#define OBJ_PHYSICAL                    0x40000000
#define OBJ_NOHEADER                    0x20000000
#define OBJ_ENCRYPTED                   0x10000000
#define OBJ_NONPERSISTENT               0x08000000

// As a cstruct enum
enum OBJ {
    VIRTUAL                             = 0x00000000,
    EPHEMERAL                           = 0x80000000,
    PHYSICAL                            = 0x40000000,
    NOHEADER                            = 0x20000000,
    ENCRYPTED                           = 0x10000000,
    NONPERSISTENT                       = 0x08000000,
};

#define MAX_CKSUM_SIZE                  8

struct obj_phys {
    char        o_cksum[MAX_CKSUM_SIZE];
    oid_t       o_oid;
    xid_t       o_xid;
    uint32_t    o_type;
    uint32_t    o_subtype;
};
typedef struct obj_phys obj_phys_t;

/*
 * Object Map
 */

/* Object Map Value Flags */
#define OMAP_VAL_DELETED                0x00000001
#define OMAP_VAL_SAVED                  0x00000002
#define OMAP_VAL_ENCRYPTED              0x00000004
#define OMAP_VAL_NOHEADER               0x00000008
#define OMAP_VAL_CRYPTO_GENERATION      0x00000010

/* Snapshot Flags */
#define OMAP_SNAPSHOT_DELETED           0x00000001
#define OMAP_SNAPSHOT_REVERTED          0x00000002

/* Object Map Flags */
#define OMAP_MANUALLY_MANAGED           0x00000001
#define OMAP_ENCRYPTING                 0x00000002
#define OMAP_DECRYPTING                 0x00000004
#define OMAP_KEYROLLING                 0x00000008
#define OMAP_CRYPTO_GENERATION          0x00000010

#define OMAP_VALID_FLAGS                0x0000001f

/* Object Map Constants */
#define OMAP_MAX_SNAP_COUNT             0xffffffff

/* Object Map Reaper Phases */
#define OMAP_REAP_PHASE_MAP_TREE        1
#define OMAP_REAP_PHASE_SNAPSHOT_TREE   2

struct omap_phys {
    obj_phys_t  om_o;
    uint32_t    om_flags;
    uint32_t    om_snap_count;
    uint32_t    om_tree_type;
    uint32_t    om_snapshot_tree_type;
    oid_t       om_tree_oid;
    oid_t       om_snapshot_tree_oid;
    xid_t       om_most_recent_snap;
    xid_t       om_pending_revert_min;
    xid_t       om_pending_revert_max;
};
typedef struct omap_phys omap_phys_t;

struct omap_key {
    oid_t       ok_oid;
    xid_t       ok_xid;
};
typedef struct omap_key omap_key_t;

struct omap_val {
    uint32_t    ov_flags;
    uint32_t    ov_size;
    paddr_t     ov_paddr;
};
typedef struct omap_val omap_val_t;

struct omap_snapshot {
    uint32_t    oms_flags;
    uint32_t    oms_pad;
    oid_t       oms_oid;
};
typedef struct omap_snapshot omap_snapshot_t;

/*
 * B-Trees
 */

/* B-Tree flags */
#define BTREE_UINT64_KEYS               0x00000001
#define BTREE_SEQUENTIAL_INSERT         0x00000002
#define BTREE_ALLOW_GHOSTS              0x00000004
#define BTREE_EPHEMERAL                 0x00000008
#define BTREE_PHYSICAL                  0x00000010
#define BTREE_NONPERSISTENT             0x00000020
#define BTREE_KV_NONALIGNED             0x00000040
#define BTREE_HASHED                    0x00000080
#define BTREE_NOHEADER                  0x00000100

// As a cstruct flag
flag BTREE {
    UINT64_KEYS                         = 0x00000001,
    SEQUENTIAL_INSERT                   = 0x00000002,
    ALLOW_GHOSTS                        = 0x00000004,
    EPHEMERAL                           = 0x00000008,
    PHYSICAL                            = 0x00000010,
    NONPERSISTENT                       = 0x00000020,
    KV_NONALIGNED                       = 0x00000040,
    HASHED                              = 0x00000080,
    NOHEADER                            = 0x00000100,
};

/* B-Tree Table of Contents Constants */
#define BTREE_TOC_ENTRY_INCREMENT       8
#define BTREE_TOC_ENTRY_MAX_UNUSED      (2 * BTREE_TOC_ENTRY_INCREMENT)

/* B-Tree Node Flags */
#define BTNODE_ROOT                     0x0001
#define BTNODE_LEAF                     0x0002

#define BTNODE_FIXED_KV_SIZE            0x0004
#define BTNODE_HASHED                   0x0008
#define BTNODE_NOHEADER                 0x0010

#define BTNODE_CHECK_KOFF_INVAL         0x8000

// As a cstruct flag
flag BTNODE {
    ROOT                                = 0x0001,
    LEAF                                = 0x0002,
    FIXED_KV_SIZE                       = 0x0004,
    HASHED                              = 0x0008,
    NOHEADER                            = 0x0010,
    CHECK_KOFF_INVAL                    = 0x8000,
};

/* B-Tree Node Constants */
#define BTREE_NODE_SIZE_DEFAULT         4096
#define BTREE_NODE_MIN_ENTRY_COUNT      4

struct nloc {
    uint16_t    off;
    uint16_t    len;
};
typedef struct nloc nloc_t;

struct btree_node_phys {
    obj_phys_t  btn_o;
    uint16_t    btn_flags;
    uint16_t    btn_level;
    uint32_t    btn_nkeys;
    nloc_t      btn_table_space;
    nloc_t      btn_free_space;
    nloc_t      btn_key_free_list;
    nloc_t      btn_val_free_list;
    // uint64_t    btn_data[];
};
typedef struct btree_node_phys btree_node_phys_t;

struct btree_info_fixed {
    uint32_t    bt_flags;
    uint32_t    bt_node_size;
    uint32_t    bt_key_size;
    uint32_t    bt_val_size;
};
typedef struct btree_info_fixed btree_info_fixed_t;

struct btree_info {
    btree_info_fixed_t  bt_fixed;
    uint32_t            bt_longest_key;
    uint32_t            bt_longest_val;
    uint64_t            bt_key_count;
    uint64_t            bt_node_count;
};
typedef struct btree_info btree_info_t;

#define BTREE_NODE_HASH_SIZE_MAX        64

struct btn_index_node_val {
    oid_t       binv_child_oid;
    char        binv_child_hash[BTREE_NODE_HASH_SIZE_MAX];
};
typedef struct btn_index_node_val btn_index_node_val_t;

#define BTOFF_INVALID                   0xffff

struct kvloc {
    nloc_t      k;
    nloc_t      v;
};
typedef struct kvloc kvloc_t;

struct kvoff {
    uint16_t    k;
    uint16_t    v;
};
typedef struct kvoff kvoff_t;

/*
 * File-System Objects
 */

struct j_key {
    uint64_t    obj_id_and_type;
};
typedef struct j_key j_key_t;

#define OBJ_ID_MASK                     0x0fffffffffffffff
#define OBJ_TYPE_MASK                   0xf000000000000000
#define OBJ_TYPE_SHIFT                  60

#define SYSTEM_OBJ_ID_MARK              0x0fffffff00000000

struct j_inode_key {
    j_key_t     hdr;
};
typedef struct j_inode_key j_inode_key_t;

struct j_inode_val {
    uint64_t    parent_id;
    uint64_t    private_id;

    uint64_t    create_time;
    uint64_t    mod_time;
    uint64_t    change_time;
    uint64_t    access_time;

    uint64_t    internal_flags;

    union {
        int32_t nchildren;
        int32_t nlink;
    };

    cp_key_class_t  default_protection_class;
    uint32_t    write_generation_counter;
    uint32_t    bsd_flags;
    uid_t       owner;
    gid_t       group;
    mode_t      mode;
    uint16_t    pad1;
    uint64_t    uncompressed_size;
    // uint8_t     xfields[];
};
typedef struct j_inode_val j_inode_val_t;

struct j_drec_key {
    j_key_t     hdr;
    uint16_t    name_len;
    char        name[name_len];
};
typedef struct j_drec_key j_drec_key_t;

#define J_DREC_LEN_MASK                 0x000003ff
#define J_DREC_HASH_MASK                0xfffffc00
#define J_DREC_HASH_SHIFT               10

struct j_drec_hashed_key {
    j_key_t     hdr;
    uint32_t    name_len_and_hash;
    char        name[name_len_and_hash & J_DREC_LEN_MASK];
};
typedef struct j_drec_hashed_key j_drec_hashed_key_t;

struct j_drec_val {
    uint64_t    file_id;
    uint64_t    date_added;
    uint16_t    flags;
    // uint8_t xfields[];
};
typedef struct j_drec_val j_drec_val_t;

struct j_dir_stats_key {
    j_key_t     hdr;
};
typedef struct j_dir_stats_key j_dir_stats_key_t;

struct j_dir_stats_val {
    uint64_t    num_children;
    uint64_t    total_size;
    uint64_t    chained_key;
    uint64_t    gen_count;
};
typedef struct j_dir_stats_val j_dir_stats_val_t;

struct j_xattr_key {
    j_key_t     hdr;
    uint16_t    name_len;
    char        name[name_len];
};
typedef struct j_xattr_key j_xattr_key_t;

struct j_xattr_val {
    uint16_t    flags;
    uint16_t    xdata_len;
    char        xdata[xdata_len];
};
typedef struct j_xattr_val j_xattr_val_t;

/* File-System Contants */
enum APFS_TYPE {
    ANY                                 = 0,
    SNAP_METADATA                       = 1,
    EXTENT                              = 2,
    INODE                               = 3,
    XATTR                               = 4,
    SIBLING_LINK                        = 5,
    DSTREAM_ID                          = 6,
    CRYPTO_STATE                        = 7,
    FILE_EXTENT                         = 8,
    DIR_REC                             = 9,
    DIR_STATS                           = 10,
    SNAP_NAME                           = 11,
    SIBLING_MAP                         = 12,
    FILE_INFO                           = 13,
    MAX_VALID                           = 13,
    MAX                                 = 15,
    INVALID                             = 15,
};
/* typedef enum j_obj_types; */

enum APFS_KIND {
    ANY                                 = 0,
    NEW                                 = 1,
    UPDATE                              = 2,
    DEAD                                = 3,
    UPDATE_REFCNT                       = 4,
    INVALID                             = 255
};
/* typedef enum j_obj_kinds; */

flag INODE {
    IS_APFS_PRIVATE                     = 0x00000001,
    MAINTAIN_DIR_STATS                  = 0x00000002,
    DIR_STATS_ORIGIN                    = 0x00000004,
    PROT_CLASS_EXPLICIT                 = 0x00000008,
    WAS_CLONED                          = 0x00000010,
    FLAG_UNUSED                         = 0x00000020,
    HAS_SECURITY_EA                     = 0x00000040,
    BEING_TRUNCATED                     = 0x00000080,
    HAS_FINDER_INFO                     = 0x00000100,
    IS_SPARSE                           = 0x00000200,
    WAS_EVER_CLONED                     = 0x00000400,
    ACTIVE_FILE_TRIMMED                 = 0x00000800,
    PINNED_TO_MAIN                      = 0x00001000,
    PINNED_TO_TIER2                     = 0x00002000,
    HAS_RSRC_FORK                       = 0x00004000,
    NO_RSRC_FORK                        = 0x00008000,
    ALLOCATION_SPILLEDOVER              = 0x00010000,
    FAST_PROMOTE                        = 0x00020000,
    HAS_UNCOMPRESSED_SIZE               = 0x00040000,
    IS_PURGEABLE                        = 0x00080000,
    WANTS_TO_BE_PURGEABLE               = 0x00100000,
    IS_SYNC_ROOT                        = 0x00200000,
    SNAPSHOT_COW_EXEMPTION              = 0x00400000,
    INHERITED_INTERNAL_FLAGS            = 0x00400002,
    CLONED_INTERNAL_FLAGS               = 0x0000c000,
};
/* typedef enum j_inode_flags */

#define APFS_INODE_PINNED_MASK          INODE.PINNED_TO_MAIN | INODE.PINNED_TO_TIER2

/* Super-user and owner changeable flags. */
#define UF_NODUMP                       0x00000001      /* do not dump file */
#define UF_IMMUTABLE                    0x00000002      /* file may not be changed */
#define UF_APPEND                       0x00000004      /* writes to file may only append */
#define UF_OPAQUE                       0x00000008      /* directory is opaque wrt. union */
#define UF_NOUNLINK                     0x00000010      /* file may not be removed or renamed */
#define UF_COMPRESSED                   0x00000020      /* file is compressed (some file-systems) */
#define UF_TRACKED                      0x00000040
#define UF_DATAVAULT                    0x00000080      /* entitlement required for reading and writing */
#define UF_HIDDEN                       0x00008000      /* hint that this item should not be displayed in a GUI */

/* Super-user changeable flags. */
#define SF_SUPPORTED                    0x009f0000      /* mask of superuser supported flags */
#define SF_SETTABLE                     0x3fff0000      /* mask of superuser changeable flags */
#define SF_SYNTHETIC                    0xc0000000      /* mask of system read-only synthetic flags */
#define SF_ARCHIVED                     0x00010000      /* file is archived */
#define SF_IMMUTABLE                    0x00020000      /* file may not be changed */
#define SF_APPEND                       0x00040000      /* writes to file may only append */
#define SF_RESTRICTED                   0x00080000      /* entitlement required for writing */
#define SF_NOUNLINK                     0x00100000      /* Item may not be removed, renamed or mounted on */
#define SF_SNAPSHOT                     0x00200000      /* snapshot inode */
#define SF_FIRMLINK                     0x00800000      /* file is a firmlink */
#define SF_DATALESS                     0x40000000      /* file is dataless object */

flag XATTR {
    DATA_STREAM                         = 0x00000001,
    DATA_EMBEDDED                       = 0x00000002,
    FILE_SYSTEM_OWNED                   = 0x00000004,
    RESERVED_8                          = 0x00000008,
    RESERVED_10                         = 0x00000010,
};
/* typedef enum j_xattr_flags; */

#define DREC_TYPE_MASK                  0x000f
// typedef enum {
//     DREC_TYPE_MASK                      = 0x000f,
//     RESERVED_10                         = 0x0010
// } dir_rec_flags;

/* Inode Numbers */
#define INVALID_INO_NUM                 0

#define ROOT_DIR_PARENT                 1
#define ROOT_DIR_INO_NUM                2
#define PRIV_DIR_INO_NUM                3
#define SNAP_DIR_INO_NUM                6
#define PURGEABLE_DIR_INO_NUM           7

#define MIN_USER_INO_NUM                16

#define UNIFIED_ID_SPACE_MARK           0x0800000000000000

/* Extended Attributes Constants */
#define XATTR_MAX_EMBEDDED_SIZE         3804
#define SYMLINK_EA_NAME                 "com.apple.fs.symlink"
#define FIRMLINK_EA_NAME                "com.apple.fs.firmlink"
#define APFS_COW_EXEMPT_COUNT_NAME      "com.apple.fs.cow-exempt-file-count"

/* File-System Object Constants */
#define OWNING_OBJ_ID_INVALID           ~0ULL
#define OWNING_OBJ_ID_UNKNOWN           ~1ULL

#define JOBJ_MAX_KEY_SIZE               832
#define JOBJ_MAX_VALUE_SIZE             3808

#define MIN_DOC_ID                      3

/* File Extent Constants */
#define FEXT_CRYPTO_ID_IS_TWEAK         0x01

/* Directory Entry File Types */
enum DT {
    UNKNOWN                             = 0,
    FIFO                                = 1,
    CHR                                 = 2,
    DIR                                 = 4,
    BLK                                 = 6,
    REG                                 = 8,
    LNK                                 = 10,
    SOCK                                = 12,
    WHT                                 = 14,
};

/*
 * Compression
 */

struct decmpfs_header {
    uint32_t    magic;
    uint32_t    algorithm;
    uint64_t    uncompressed_size;
};

#define DECMPFS_MAGIC                   b"cmpf"
#define DECMPFS_BLOCK_SIZE              0x10000

#define DECMPFS_ZLIB_ATTR               3
#define DECMPFS_ZLIB_RSRC               4
#define DECMPFS_LZVN_ATTR               7
#define DECMPFS_LZVN_RSRC               8
#define DECMPFS_PLAIN_ATTR              9
#define DECMPFS_PLAIN_RSRC              10
#define DECMPFS_LZFSE_ATTR              11
#define DECMPFS_LZFSE_RSRC              12
#define DECMPFS_LZBITMAP_ATTR           13
#define DECMPFS_LZBITMAP_RSRC           14

/*
 * Data Streams
 */

struct j_phys_ext_key {
    j_key_t     hdr;
};
typedef struct j_phys_ext_key j_phys_ext_key_t;

struct j_phys_ext_val {
    uint64_t    len_and_kind;
    uint64_t    owning_obj_id;
    int32_t     refcnt;
};
typedef struct j_phys_ext_val j_phys_ext_val_t;

#define PEXT_LEN_MASK                   0x0fffffffffffffff
#define PEXT_KIND_MASK                  0xf000000000000000
#define PEXT_KIND_SHIFT                 60

struct j_file_extent_key {
    j_key_t     hdr;
    uint64_t    logical_addr;
};
typedef struct j_file_extent_key j_file_extent_key_t;

struct j_file_extent_val {
    uint64_t    len_and_flags;
    uint64_t    phys_block_num;
    uint64_t    crypto_id;
};
typedef struct j_file_extent_val j_file_extent_val_t;

#define J_FILE_EXTENT_LEN_MASK          0x00ffffffffffffff
#define J_FILE_EXTENT_FLAG_MASK         0xff00000000000000
#define J_FILE_EXTENT_FLAG_SHIFT        56

struct j_dstream_id_key {
    j_key_t     hdr;
};
typedef struct j_dstream_id_key j_dstream_id_key_t;

struct j_dstream_id_val {
    uint32_t    refcnt;
};
typedef struct j_dstream_id_val j_dstream_id_val_t;

struct j_dstream {
    uint64_t    size;
    uint64_t    alloced_size;
    uint64_t    default_crypto_id;
    uint64_t    total_bytes_written;
    uint64_t    total_bytes_read;
};
typedef struct j_dstream j_dstream_t;

struct j_xattr_dstream {
    uint64_t    xattr_obj_id;
    j_dstream_t dstream;
};
typedef struct j_xattr_dstream j_xattr_dstream_t;

/*
 * Extended fields
 */

struct x_field {
    uint8_t     x_type;
    uint8_t     x_flags;
    uint16_t    x_size;
};
typedef struct x_field x_field_t;

struct xf_blob {
    uint16_t    xf_num_exts;
    uint16_t    xf_used_data;
    x_field_t   xf_exts[xf_num_exts];
    char        xf_data[xf_used_data];
};
typedef struct xf_blob xf_blob_t;

/* Extended Field Types */
enum DREC_EXT_TYPE {
    SIBLING_ID                          = 1,
};

enum INO_EXT_TYPE {
    SNAP_XID                            = 1,
    DELTA_TREE_OID                      = 2,
    DOCUMENT_ID                         = 3,
    NAME                                = 4,
    PREV_FSIZE                          = 5,
    RESERVED_6                          = 6,
    FINDER_INFO                         = 7,
    DSTREAM                             = 8,
    RESERVED_9                          = 9,
    DIR_STATS_KEY                       = 10,
    FS_UUID                             = 11,
    RESERVED_12                         = 12,
    SPARSE_BYTES                        = 13,
    RDEV                                = 14,
    PURGEABLE_FLAGS                     = 15,
    ORIG_SYNC_ROOT_ID                   = 16,
};

/* Extended Field Flags */
flag XF {
    DATA_DEPENDENT                      = 0x0001,
    DO_NOT_COPY                         = 0x0002,
    RESERVED_4                          = 0x0004,
    CHILDREN_INHERIT                    = 0x0008,
    USER_FIELD                          = 0x0010,
    SYSTEM_FIELD                        = 0x0020,
    RESERVED_40                         = 0x0040,
    RESERVED_80                         = 0x0080,
};

/*
 * Siblings
 */

struct j_sibling_key {
    j_key_t     hdr;
    uint64_t    sibling_id;
};
typedef struct j_sibling_key j_sibling_key_t;

struct j_sibling_val {
    uint64_t    parent_id;
    uint16_t    name_len;
    char        name[name_len];
};
typedef struct j_sibling_val j_sibling_val_t;

struct j_sibling_map_key {
    j_key_t     hdr;
};
typedef struct j_sibling_map_key j_sibling_map_key_t;

struct j_sibling_map_val {
    uint64_t    file_id;
};
typedef struct j_sibling_map_val j_sibling_map_val_t;

/*
 * Snapshot Metadata
 */

struct j_snap_metadata_key {
    j_key_t     hdr;
};
typedef struct j_snap_metadata_key j_snap_metadata_key_t;

struct j_snap_metadata_val {
    oid_t       extentref_tree_oid;
    oid_t       sblock_oid;
    uint64_t    create_time;
    uint64_t    change_time;
    uint64_t    inum;
    uint32_t    extentref_tree_type;
    uint32_t    flags;
    uint16_t    name_len;
    char        name[name_len];
};
typedef struct j_snap_metadata_val j_snap_metadata_val_t;

struct j_snap_name_key {
    j_key_t     hdr;
    uint16_t    name_len;
    char        name[name_len];
};
typedef struct j_snap_name_key j_snap_name_key_t;

struct j_snap_name_val {
    xid_t       snap_xid;
};
typedef struct j_snap_name_val j_snap_name_val_t;

enum SNAP_META {
    PENDING_DATALESS                    = 0x00000001,
    MERGE_IN_PROGRESS                   = 0x00000002,
};
/* typedef enum snap_meta_flags; */

typedef struct snap_meta_ext {
    uint32_t    sme_version;
    uint32_t    sme_flags;
    xid_t       sme_snap_xid;
    char        sme_uuid[16];
    uint64_t    sme_token;
};
typedef struct snap_meta_ext snap_meta_ext_t;

struct snap_meta_ext_obj_phys {
    obj_phys_t      smeop_o;
    snap_meta_ext_t smeop_sme;
};
typedef struct snap_meta_ext_obj_phys snap_meta_ext_obj_phys_t;

/*
 * Encryption
 */

#define CRYPTO_SW_ID                    4
#define CRYPTO_RESERVED_5               5
#define APFS_UNASSIGNED_CRYPTO_ID       (~0)

struct keybag_entry {
    char        ke_uuid[16];
    uint16_t    ke_tag;
    uint16_t    ke_keylen;
    uint8_t     padding[4];
    char        ke_keydata[0];
};
typedef struct keybag_entry keybag_entry_t;

#define APFS_VOL_KEYBAG_ENTRY_MAX_SIZE  512
#define APFS_FV_PERSONAL_RECOVERY_KEY_UUID  "EBC6C064-0000-11AA-AA11-00306543ECAC"

struct kb_locker {
    uint16_t        kl_version;
    uint16_t        kl_nkeys;
    uint32_t        kl_nbytes;
    uint8_t         padding[8];
    keybag_entry_t  kl_entries[0];
};
typedef struct kb_locker kb_locker_t;

#define APFS_KEYBAG_VERSION             2

struct media_keybag {
    obj_phys_t  mk_obj;
    kb_locker_t mk_locker;
};
typedef struct media_keybag media_keybag_t;

enum KB_TAG {
    UNKNOWN                             = 0,
    RESERVED_1                          = 1,
    VOLUME_KEY                          = 2,
    VOLUME_UNLOCK_RECORDS               = 3,
    VOLUME_PASSPHRASE_HINT              = 4,
    WRAPPING_M_KEY                      = 5,
    KB_TAG_VOLUME_M_KEY                 = 6,
    KB_TAG_RESERVED_F8                  = 0xF8,
};

enum PROTECTION_CLASS {
    DIR_NONE                            = 0,
    A                                   = 1,
    B                                   = 2,
    C                                   = 3,
    D                                   = 4,
    F                                   = 6,
    M                                   = 14,
};

#define CP_EFFECTIVE_CLASSMASK          0x0000001f

struct wrapped_meta_crypto_state {
    uint16_t            major_version;
    uint16_t            minor_version;
    crypto_flags_t      cpflags;
    cp_key_class_t      persistent_class;
    cp_key_os_version_t key_os_version;
    cp_key_revision_t   key_revision;
    uint16_t            unused;
};
typedef struct wrapped_meta_crypto_state wrapped_meta_crypto_state_t;

struct wrapped_crypto_state {
    uint16_t            major_version;
    uint16_t            minor_version;
    crypto_flags_t      cpflags;
    cp_key_class_t      persistent_class;
    cp_key_os_version_t key_os_version;
    cp_key_revision_t   key_revision;
    uint16_t            key_len;
    char                persistent_key[key_len];
};
typedef struct wrapped_crypto_state wrapped_crypto_state_t;

struct j_crypto_key {
    j_key_t     hdr;
};
typedef struct j_crypto_key j_crypto_key_t;

struct j_crypto_val {
    uint32_t    refcnt;
    wrapped_crypto_state_t  state;
};
typedef struct j_crypto_val j_crypto_val_t;

/*
 * Encryption Rolling
 */

enum ER_PHASE {
    OMAP_ROLL                           = 1,
    DATA_ROLL                           = 2,
    SNAP_ROLL                           = 3,
};

enum {
    ER_512B_BLOCKSIZE                   = 0,
    ER_2KiB_BLOCKSIZE                   = 1,
    ER_4KiB_BLOCKSIZE                   = 2,
    ER_8KiB_BLOCKSIZE                   = 3,
    ER_16KiB_BLOCKSIZE                  = 4,
    ER_32KiB_BLOCKSIZE                  = 5,
    ER_64KiB_BLOCKSIZE                  = 6,
};

/* Encryption Rolling Flags */
#define ERSB_FLAG_ENCRYPTING            0x00000001
#define ERSB_FLAG_DECRYPTING            0x00000002
#define ERSB_FLAG_KEYROLLING            0x00000004
#define ERSB_FLAG_PAUSED                0x00000008
#define ERSB_FLAG_FAILED                0x00000010
#define ERSB_FLAG_CID_IS_TWEAK          0x00000020
#define ERSB_FLAG_FREE_1                0x00000040
#define ERSB_FLAG_FREE_2                0x00000080

#define ERSB_FLAG_CM_BLOCK_SIZE_MASK    0x00000F00
#define ERSB_FLAG_CM_BLOCK_SIZE_SHIFT   8

#define ERSB_FLAG_ER_PHASE_MASK         0x00003000
#define ERSB_FLAG_ER_PHASE_SHIFT        12
#define ERSB_FLAG_FROM_ONEKEY           0x00004000

/* Encryption Rolling Constants */
#define ER_CHECKSUM_LENGTH              8
#define ER_MAGIC                        b'FLAB'
#define ER_VERSION                      1

#define ER_MAX_CHECKSUM_COUNT_SHIFT     16
#define ER_CUR_CHECKSUM_COUNT_MASK      0x0000FFFF

struct er_state_phys_header {
    obj_phys_t  ersb_o;
    uint32_t    ersb_magic;
    uint32_t    ersb_version;
};
typedef struct er_state_phys_header er_state_phys_header_t;

struct er_state_phys {
    er_state_phys_header_t  ersb_header;
    uint64_t    ersb_flags;
    uint64_t    ersb_snap_xid;
    uint64_t    ersb_current_fext_obj_id;
    uint64_t    ersb_file_offset;
    uint64_t    ersb_progress;
    uint64_t    ersb_total_blk_to_encrypt;
    oid_t       ersb_blockmap_oid;
    uint64_t    ersb_tidemark_obj_id;
    uint64_t    ersb_recovery_extents_count;
    oid_t       ersb_recovery_list_oid;
    uint64_t    ersb_recovery_length;
};
typedef struct er_state_phys er_state_phys_t;

struct er_state_phys_v1 {
    er_state_phys_header_t  ersb_header;
    uint64_t    ersb_flags;
    uint64_t    ersb_snap_xid;
    uint64_t    ersb_current_fext_obj_id;
    uint64_t    ersb_file_offset;
    uint64_t    ersb_fext_pbn;
    uint64_t    ersb_paddr;
    uint64_t    ersb_progress;
    uint64_t    ersb_total_blk_to_encrypt;
    uint64_t    ersb_blockmap_oid;
    uint32_t    ersb_checksum_count;
    uint32_t    ersb_reserved;
    uint64_t    ersb_fext_cid;
    uint8_t     ersb_checksum[0];
};
typedef struct er_state_phys_v1 er_state_phys_v1_t;

struct er_recovery_block_phys {
    obj_phys_t  erb_o;
    uint64_t    erb_offset;
    oid_t       erb_next_oid;
    char        erb_data[0];
};
typedef struct er_recovery_block_phys er_recovery_block_phys_t;

struct gbitmap_block_phys {
    obj_phys_t  bmb_o;
    uint64_t    bmb_field[0];
};
typedef struct gbitmap_block_phys gbitmap_block_phys_t;

struct gbitmap_phys {
    obj_phys_t  bm_o;
    oid_t       bm_tree_oid;
    uint64_t    bm_bit_count;
    uint64_t    bm_flags;
};
typedef struct gbitmap_phys gbitmap_phys_t;

/*
 * EFI Jumpstart
 */

#define NX_EFI_JUMPSTART_MAGIC          b'RDSJ'
#define NX_EFI_JUMPSTART_VERSION        1

struct nx_efi_jumpstart {
    obj_phys_t  nej_o;
    uint32_t    nej_magic;
    uint32_t    nej_version;
    uint32_t    nej_efi_file_len;
    uint32_t    nej_num_extents;
    uint64_t    nej_reserved[16];
    prange_t    nej_rec_extents[];
};
typedef struct nx_efi_jumpstart nx_efi_jumpstart_t;

/*
 * Container
 */

#define NX_MAGIC                        b'BSXN'
#define NX_MAX_FILE_SYSTEMS             100
#define NX_EPH_INFO_COUNT               4
#define NX_EPH_MIN_BLOCK_COUNT          8
#define NX_MAX_FILE_SYSTEM_EPH_STRUCTS  4
#define NX_TX_MIN_CHECKPOINT_COUNT      4
#define NX_EPH_INFO_VERSION_1           1

#define NX_MINIMUM_BLOCK_SIZE           4096
#define NX_DEFAULT_BLOCK_SIZE           4096
#define NX_MAXIMUM_BLOCK_SIZE           65536
#define NX_MINIMUM_CONTAINER_SIZE       1048576

#define NX_RESERVED_1                   0x00000001
#define NX_RESERVED_2                   0x00000002
#define NX_CRYPTO_SW                    0x00000004

flag NX_FEATURE {
    DEFRAG                              = 0x0000000000000001,
    LCFD                                = 0x0000000000000002,
};

flag NX_INCOMPAT {
    VERSION1                            = 0x0000000000000001,
    VERSION2                            = 0x0000000000000002,
    FUSION                              = 0x0000000000000100,
};

enum {
    NX_CNTR_OBJ_CKSUM_SET               = 0,
    NX_CNTR_OBJ_CKSUM_FAIL              = 1,
    NX_NUM_COUNTERS                     = 32,
};

struct nx_superblock {
    obj_phys_t  nx_o;
    uint32_t    nx_magic;
    uint32_t    nx_block_size;
    uint64_t    nx_block_count;
    uint64_t    nx_features;
    uint64_t    nx_readonly_compatible_features;
    uint64_t    nx_incompatible_features;
    char        nx_uuid[16];
    oid_t       nx_next_oid;
    xid_t       nx_next_xid;
    uint32_t    nx_xp_desc_blocks;
    uint32_t    nx_xp_data_blocks;
    paddr_t     nx_xp_desc_base;
    paddr_t     nx_xp_data_base;
    uint32_t    nx_xp_desc_next;
    uint32_t    nx_xp_data_next;
    uint32_t    nx_xp_desc_index;
    uint32_t    nx_xp_desc_len;
    uint32_t    nx_xp_data_index;
    uint32_t    nx_xp_data_len;
    oid_t       nx_spaceman_oid;
    oid_t       nx_omap_oid;
    oid_t       nx_reaper_oid;
    uint32_t    nx_test_type;
    uint32_t    nx_max_file_systems;
    oid_t       nx_fs_oid[NX_MAX_FILE_SYSTEMS];
    uint64_t    nx_counters[NX_NUM_COUNTERS];
    prange_t    nx_blocked_out_prange;
    oid_t       nx_evict_mapping_tree_oid;
    uint64_t    nx_flags;
    paddr_t     nx_efi_jumpstart;
    char        nx_fusion_uuid[16];
    prange_t    nx_keylocker;
    uint64_t    nx_ephemeral_info[NX_EPH_INFO_COUNT];
    oid_t       nx_test_oid;
    oid_t       nx_fusion_mt_oid;
    oid_t       nx_fusion_wbc_oid;
    prange_t    nx_fusion_wbc;
    uint64_t    nx_newest_mounted_version;
    prange_t    nx_mkb_locker;
};

#define CHECKPOINT_MAP_LAST             0x00000001

struct checkpoint_mapping {
    uint32_t    cpm_type;
    uint32_t    cpm_subtype;
    uint32_t    cpm_size;
    uint32_t    cpm_pad;
    oid_t       cpm_fs_oid;
    oid_t       cpm_oid;
    oid_t       cpm_paddr;
};
typedef struct checkpoint_mapping checkpoint_mapping_t;

struct checkpoint_map_phys {
    obj_phys_t              cpm_o;
    uint32_t                cpm_flags;
    uint32_t                cpm_count;
    checkpoint_mapping_t    cpm_map[cpm_count];
};
typedef struct checkpoint_map_phys checkpoint_map_phys_t;

struct evict_mapping_val {
    paddr_t     dst_paddr;
    uint64_t    len;
};
typedef struct evict_mapping_val evict_mapping_val_t;

/*
 * Volume
 */

#define APFS_MAGIC                      b'BSPA'
#define APFS_MAX_HIST                   8
#define APFS_VOLNAME_LEN                256

#define APFS_MODIFIED_NAMELEN           32

struct apfs_modified_by {
    char        id[APFS_MODIFIED_NAMELEN];
    uint64_t    timestamp;
    xid_t       last_xid;
};
typedef struct apfs_modified_by apfs_modified_by_t;

struct apfs_superblock {
    obj_phys_t  apfs_o;
    uint32_t    apfs_magic;
    uint32_t    apfs_fs_index;
    uint64_t    apfs_features;
    uint64_t    apfs_readonly_compatible_features;
    uint64_t    apfs_incompatible_features;
    uint64_t    apfs_unmount_time;
    uint64_t    apfs_fs_reserve_block_count;
    uint64_t    apfs_fs_quota_block_count;
    uint64_t    apfs_fs_alloc_count;
    wrapped_meta_crypto_state_t apfs_meta_crypto;
    uint32_t    apfs_root_tree_type;
    uint32_t    apfs_extentref_tree_type;
    uint32_t    apfs_snap_meta_tree_type;
    oid_t       apfs_omap_oid;
    oid_t       apfs_root_tree_oid;
    oid_t       apfs_extentref_tree_oid;
    oid_t       apfs_snap_meta_tree_oid;
    xid_t       apfs_revert_to_xid;
    oid_t       apfs_revert_to_sblock_oid;
    uint64_t    apfs_next_obj_id;
    uint64_t    apfs_num_files;
    uint64_t    apfs_num_directories;
    uint64_t    apfs_num_symlinks;
    uint64_t    apfs_num_other_fsobjects;
    uint64_t    apfs_num_snapshots;
    uint64_t    apfs_total_blocks_alloced;
    uint64_t    apfs_total_blocks_freed;
    char        apfs_vol_uuid[16];
    uint64_t    apfs_last_mod_time;
    uint64_t    apfs_fs_flags;
    apfs_modified_by_t  apfs_formatted_by;
    apfs_modified_by_t  apfs_modified_by[APFS_MAX_HIST];
    char        apfs_volname[APFS_VOLNAME_LEN];
    uint32_t    apfs_next_doc_id;
    uint16_t    apfs_role;
    uint16_t    reserved;
    xid_t       apfs_root_to_xid;
    oid_t       apfs_er_state_oid;
    uint64_t    apfs_cloneinfo_id_epoch;
    xid_t       apfs_cloneinfo_xid;
    oid_t       apfs_snap_meta_ext_oid;
    char        apfs_volume_group_id[16];
    oid_t       apfs_integrity_meta_oid;
    oid_t       apfs_fext_tree_oid;
    uint32_t    apfs_fext_tree_type;
    uint32_t    apfs_pfkur_tree_type;
    oid_t       apfs_pfkur_tree_oid;
    // Undocumented from here on
    xid_t       apfs_doc_id_index_xid;
    uint32_t    apfs_doc_id_index_flags;
    uint32_t    apfs_doc_id_tree_type;
    oid_t       apfs_doc_id_tree_oid;
    oid_t       apfs_prev_doc_id_tree_oid;
    uint64_t    apfs_doc_id_fixup_cursor;
    oid_t       apfs_sec_root_tree_oid;
    uint32_t    apfs_sec_root_tree_type;
    uint32_t    apfs_clone_group_tree_flags;
};
typedef struct apfs_superblock apfs_superblock_t;

flag APFS_FEATURE {
    DEFRAG_PRERELEASE                   = 0x00000001,
    HARDLINK_MAP_RECORDS                = 0x00000002,
    DEFRAG                              = 0x00000004,
    STRICTATIME                         = 0x00000008,
    VOLGRP_SYSTEM_INO_SPACE             = 0x00000010,
};

flag APFS_INCOMPAT {
    CASE_INSENSITIVE                    = 0x00000001,
    DATALESS_SNAPS                      = 0x00000002,
    ENC_ROLLED                          = 0x00000004,
    NORMALIZATION_INSENSITIVE           = 0x00000008,
    INCOMPLETE_RESTORE                  = 0x00000010,
    SEALED_VOLUME                       = 0x00000020,
    PFK                                 = 0x00000040,   // Undocumented name
    RESERVED_80                         = 0x00000080,   // Maybe EXTENT_PREALLOC according to linux-apfs-rw
    SECONDARY_FSROOT                    = 0x00000100,   // Undocumented name
};

#define APFS_VOLUME_ENUM_SHIFT          6

flag APFS_FS {
    UNENCRYPTED                         = 0x00000001,
    RESERVED_2                          = 0x00000002,
    RESERVED_4                          = 0x00000004,
    ONEKEY                              = 0x00000008,
    SPILLEDOVER                         = 0x00000010,
    RUN_SPILLOVER_CLEANER               = 0x00000020,
    ALWAYS_CHECK_EXTENTREF              = 0x00000040,
    PREVIOUSLY_SEALED                   = 0x00000080,   // Undocumented name
    PFK                                 = 0x00000100,   // Undocumented name
    RESERVED_200                        = 0x00000200,
    RESERVED_400                        = 0x00000400,
    RESERVED_800                        = 0x00000800,
};

enum APFS_VOL_ROLE {
    NONE                                = 0x0000,
    SYSTEM                              = 0x0001,
    USER                                = 0x0002,
    RECOVERY                            = 0x0004,
    VM                                  = 0x0008,
    PREBOOT                             = 0x0010,
    INSTALLER                           = 0x0020,
    DATA                                = (1 << APFS_VOLUME_ENUM_SHIFT),
    BASEBAND                            = (2 << APFS_VOLUME_ENUM_SHIFT),
    UPDATE                              = (3 << APFS_VOLUME_ENUM_SHIFT),
    XART                                = (4 << APFS_VOLUME_ENUM_SHIFT),
    HARDWARE                            = (5 << APFS_VOLUME_ENUM_SHIFT),
    BACKUP                              = (6 << APFS_VOLUME_ENUM_SHIFT),
    RESERVED_7                          = (7 << APFS_VOLUME_ENUM_SHIFT),
    RESERVED_8                          = (8 << APFS_VOLUME_ENUM_SHIFT),
    ENTERPRISE                          = (9 << APFS_VOLUME_ENUM_SHIFT),
    RESERVED_10                         = (10 << APFS_VOLUME_ENUM_SHIFT),
    PRELOGIN                            = (11 << APFS_VOLUME_ENUM_SHIFT),
};

/*
 * Sealed Volumes
 */

enum apfs_hash_type_t : uint32_t {
    APFS_HASH_INVALID                   = 0,
    APFS_HASH_SHA256                    = 0x1,
    APFS_HASH_SHA512_256                = 0x2,
    APFS_HASH_SHA384                    = 0x3,
    APFS_HASH_SHA512                    = 0x4,
};

#define APFS_HASH_CCSHA256_SIZE         32
#define APFS_HASH_CCSHA512_256_SIZE     32
#define APFS_HASH_CCSHA384_SIZE         48
#define APFS_HASH_CCSHA512_SIZE         64

#define APFS_HASH_MAX_SIZE              64

struct integrity_meta_phys {
    obj_phys_t          im_o;
    uint32_t            im_version;
    uint32_t            im_flags;
    apfs_hash_type_t    im_hash_type;
    uint32_t            im_root_hash_offset;
    xid_t               im_broken_xid;
    uint64_t            im_reserved[9];
};
typedef struct integrity_meta_phys integrity_meta_phys_t;

/* Integrity Metadata Flags */
#define APFS_SEAL_BROKEN                (1 << 0)

struct fext_tree_key {
    uint64_t    private_id;
    uint64_t    logical_addr;
};
typedef struct fext_tree_key fext_tree_key_t;

struct fext_tree_val {
    uint64_t    len_and_flags;
    uint64_t    phys_block_num;
};
typedef struct fext_tree_val fext_tree_val_t;

struct j_file_info_key {
    j_key_t     hdr;
    uint64_t    info_and_lba;
};
typedef struct j_key_t j_file_info_key_t;

#define J_FILE_INFO_LBA_MASK            0x00ffffffffffffff
#define J_FILE_INFO_TYPE_MASK           0xff00000000000000
#define J_FILE_INFO_TYPE_SHIFT          56

struct j_file_data_hash_val {
    uint16_t    hashed_len;
    uint8_t     hash_size;
    char        hash[hash_size];
};
typedef struct j_file_data_hash_val j_file_data_hash_val_t;

struct j_file_info_val {
    union {
        j_file_data_hash_val_t dhash;
    };
};
typedef struct j_file_data_hash_val_t j_file_info_val_t;

enum APFS_FILE_INFO {
    DATA_HASH                           = 1,
};
/* typedef enum j_obj_file_info_type; */

/*
 * Space Manager
 */

typedef uint64_t spaceman_free_queue_val_t;

#define SM_ALLOCZONE_INVALID_END_BOUNDARY       0
#define SM_ALLOCZONE_NUM_PREVIOUS_BOUNDARIES    7

#define SM_DATAZONE_ALLOCZONE_COUNT     8

#define SM_FLAG_VERSIONED               0x00000001

enum {
    SFQ_IP                              = 0,
    SFQ_MAIN                            = 1,
    SFQ_TIER2                           = 2,
    SFQ_COUNT                           = 3
};

enum {
    SD_MAIN                             = 0,
    SD_TIER2                            = 1,
    SD_COUNT                            = 2
};

/* Chunk Info Block Constants */
#define CI_COUNT_MASK                   0x000fffff
#define CI_COUNT_RESERVED_MASK          0xfff00000

/* Internal-Pool Bitmap */
#define SPACEMAN_IP_BM_TX_MULTIPLIER    16
#define SPACEMAN_IP_BM_INDEX_INVALID    0xffff
#define SPACEMAN_IP_BM_BLOCK_COUNT_MAX  0xfffe

struct chunk_info {
    uint64_t    ci_xid;
    uint64_t    ci_addr;
    uint32_t    ci_block_count;
    uint32_t    ci_free_count;
    paddr_t     ci_bitmap_addr;
};
typedef struct chunk_info chunk_info_t;

struct chunk_info_block {
    obj_phys_t      cib_o;
    uint32_t        cib_index;
    uint32_t        cib_chunk_info_count;
    chunk_info_t    cib_chunk_info[cib_chunk_info_count];
};
typedef struct chunk_info_block chunk_info_block_t;

struct cib_addr_block {
    obj_phys_t  cab_o;
    uint32_t    cab_index;
    uint32_t    cab_cib_count;
    paddr_t     cab_cib_addr[cab_cib_count];
};
typedef struct cib_addr_block cib_addr_block_t;

struct spaceman_free_queue_key {
    xid_t       sfqk_xid;
    paddr_t     sfqk_paddr;
};
typedef struct spaceman_free_queue_key spaceman_free_queue_key_t;

struct spaceman_free_queue_entry {
    spaceman_free_queue_key_t   sfqe_key;
    spaceman_free_queue_val_t   sfqe_count;
};
typedef struct spaceman_free_queue_entry spaceman_free_queue_entry_t;

struct spaceman_free_queue {
    uint64_t    sfq_count;
    oid_t       sfq_tree_oid;
    xid_t       sfq_oldest_xid;
    uint16_t    sfq_tree_node_limit;
    uint16_t    sfq_pad16;
    uint32_t    sfq_pad32;
    uint64_t    sfq_reserved;
};
typedef struct spaceman_free_queue spaceman_free_queue_t;

struct spaceman_device {
    uint64_t    sm_block_count;
    uint64_t    sm_chunk_count;
    uint32_t    sm_cib_count;
    uint32_t    sm_cab_count;
    uint64_t    sm_free_count;
    uint32_t    sm_addr_offset;
    uint32_t    sm_reserved;
    uint64_t    sm_reserved2;
};
typedef struct spaceman_device spaceman_device_t;

struct spaceman_allocation_zone_boundaries {
    uint64_t    saz_zone_start;
    uint64_t    saz_zone_end;
};
typedef struct spaceman_allocation_zone_boundaries spaceman_allocation_zone_boundaries_t;

struct spaceman_allocation_zone_info_phys {
    spaceman_allocation_zone_boundaries_t   saz_current_boundaries;
    spaceman_allocation_zone_boundaries_t   saz_previous_boundaries[SM_ALLOCZONE_NUM_PREVIOUS_BOUNDARIES];
    uint16_t    saz_zone_id;
    uint16_t    saz_previous_boundary_index;
    uint32_t    saz_reserved;
};
typedef struct spaceman_allocation_zone_info_phys spaceman_allocation_zone_info_phys_t;

struct spaceman_datazone_info_phys {
    spaceman_allocation_zone_info_phys_t    sdz_allocation_zones[SD_COUNT][SM_DATAZONE_ALLOCZONE_COUNT];
};
typedef struct spaceman_datazone_info_phys spaceman_datazone_info_phys_t;

struct spaceman_phys {
    obj_phys_t  sm_o;
    uint32_t    sm_block_size;
    uint32_t    sm_blocks_per_chunk;
    uint32_t    sm_chunks_per_cib;
    uint32_t    sm_cibs_per_cab;
    spaceman_device_t   sm_dev[SD_COUNT];
    uint32_t    sm_flags;
    uint32_t    sm_ip_bm_tx_multiplier;
    uint64_t    sm_ip_block_count;
    uint32_t    sm_ip_bm_size_in_blocks;
    uint32_t    sm_ip_bm_block_count;
    paddr_t     sm_ip_bm_base;
    paddr_t     sm_ip_base;
    uint64_t    sm_fs_reserve_block_count;
    uint64_t    sm_fs_reserve_alloc_count;
    spaceman_free_queue_t   sm_fq[SFQ_COUNT];
    uint16_t    sm_ip_bm_free_head;
    uint16_t    sm_ip_bm_free_tail;
    uint32_t    sm_ip_bm_xid_offset;
    uint32_t    sm_ip_bitmap_offset;
    uint32_t    sm_ip_bm_free_next_offset;
    uint32_t    sm_version;
    uint32_t    sm_struct_size;
    spaceman_datazone_info_phys_t   sm_datazone;
};
typedef struct spaceman_phys spaceman_phys_t;

/*
 * Reaper
 */

/* Volume Reaper States */
enum APFS_REAP_PHASE {
    START                               = 0,
    SNAPSHOTS                           = 1,
    ACTIVE_FS                           = 2,
    DESTROY_OMAP                        = 3,
    DONE                                = 4
};

/* Reaper Flags */
#define NR_BHM_FLAG                     0x00000001
#define NR_CONTINUE                     0x00000002

/* Reaper List Entry Flags */
#define NRLE_VALID                      0x00000001
#define NRLE_REAP_ID_RECORD             0x00000002
#define NRLE_CALL                       0x00000004
#define NRLE_COMPLETION                 0x00000008
#define NRLE_CLEANUP                    0x00000010

/* Reaper List Flags */
#define NRL_INDEX_INVALID               0xffffffff

struct nx_reaper_phys {
    obj_phys_t  nr_o;
    uint64_t    nr_next_reap_id;
    uint64_t    nr_completed_id;
    oid_t       nr_head;
    oid_t       nr_tail;
    uint32_t    nr_flags;
    uint32_t    nr_rlcount;
    uint32_t    nr_type;
    uint32_t    nr_size;
    oid_t       nr_fs_oid;
    oid_t       nr_oid;
    xid_t       nr_xid;
    uint32_t    nr_nrle_flags;
    uint32_t    nr_state_buffer_size;
    uint8_t     nr_state_buffer[];
};
typedef struct nx_reaper_phys nx_reaper_phys_t;

struct nx_reap_list_entry {
    uint32_t    nrle_next;
    uint32_t    nrle_flags;
    uint32_t    nrle_type;
    uint32_t    nrle_size;
    oid_t       nrle_fs_oid;
    oid_t       nrle_oid;
    xid_t       nrle_xid;
};
typedef struct nx_reap_list_entry nx_reap_list_entry_t;

struct nx_reap_list_phys {
    obj_phys_t  nrl_o;
    oid_t       nrl_next;
    uint32_t    nrl_flags;
    uint32_t    nrl_max;
    uint32_t    nrl_count;
    uint32_t    nrl_first;
    uint32_t    nrl_last;
    uint32_t    nrl_free;
    // nx_reap_list_entry_t    nrl_entries[];
};
typedef struct nx_reap_list_phys nx_reap_list_phys_t;

struct omap_reap_state {
    uint32_t    omr_phase;
    omap_key_t  omr_ok;
};
typedef struct omap_reap_state omap_reap_state_t;

struct omap_cleanup_state {
    uint32_t    omc_cleaning;
    uint32_t    omc_omsflags;
    xid_t       omc_sxidprev;
    xid_t       omc_sxidstart;
    xid_t       omc_sxidend;
    xid_t       omc_sxidnext;
    omap_key_t  omc_curkey;
};
typedef struct omap_cleanup_state omap_cleanup_state_t;

struct apfs_reap_state {
    uint64_t    last_pbn;
    xid_t       cur_snap_xid;
    uint32_t    phase;
};
typedef struct apfs_reap_state apfs_reap_state_t;

/*
 * Fusion
 */

typedef paddr_t fusion_mt_key_t;

#define FUSION_TIER2_DEVICE_BYTE_ADDR   0x4000000000000000

struct fusion_wbc_phys {
    obj_phys_t  fwp_objHdr;
    uint64_t    fwp_version;
    oid_t       fwp_listHeadOid;
    oid_t       fwp_listTailOid;
    uint64_t    fwp_stableHeadOffset;
    uint64_t    fwp_stableTailOffset;
    uint32_t    fwp_listBlocksCount;
    uint32_t    fwp_reserved;
    uint64_t    fwp_usedByRC;
    prange_t    fwp_rcStash;
};
typedef struct fusion_wbc_phys fusion_wbc_phys_t;

struct fusion_wbc_list_entry {
    paddr_t     fwle_wbcLba;
    paddr_t     fwle_targetLba;
    uint64_t    fwle_length;
};
typedef struct fusion_wbc_list_entry fusion_wbc_list_entry_t;

struct fusion_wbc_list_phys {
    obj_phys_t  fwlp_objHdr;
    uint64_t    fwlp_version;
    uint64_t    fwlp_tailOffset;
    uint32_t    fwlp_indexBegin;
    uint32_t    fwlp_indexEnd;
    uint32_t    fwlp_indexMax;
    uint32_t    fwlp_reserved;
    fusion_wbc_list_entry_t     fwlp_listEntries[];
};
typedef struct fusion_wbc_list_phys fusion_wbc_list_phys_t;

struct fusion_mt_val {
    paddr_t     fmv_lba;
    uint32_t    fmv_length;
    uint32_t    fmv_flags;
};
typedef struct fusion_mt_val fusion_mt_val_t;

#define FUSION_MT_DIRTY                 (1 << 0)
#define FUSION_MT_TENANT                (1 << 1)
#define FUSION_MT_ALLFLAGS              (FUSION_MT_DIRTY | FUSION_MT_TENANT)
"""
c_apfs = cstruct()
c_apfs.load(apfs_def)

FILESYSTEM_OBJECT_TYPE_MAP = {
    c_apfs.APFS_TYPE.SNAP_METADATA: (c_apfs.j_snap_metadata_key, c_apfs.j_snap_metadata_val),
    c_apfs.APFS_TYPE.EXTENT: (c_apfs.j_phys_ext_key, c_apfs.j_phys_ext_val),
    c_apfs.APFS_TYPE.INODE: (c_apfs.j_inode_key, c_apfs.j_inode_val),
    c_apfs.APFS_TYPE.XATTR: (c_apfs.j_xattr_key, c_apfs.j_xattr_val),
    c_apfs.APFS_TYPE.SIBLING_LINK: (c_apfs.j_sibling_key, c_apfs.j_sibling_val),
    c_apfs.APFS_TYPE.DSTREAM_ID: (c_apfs.j_dstream_id_key, c_apfs.j_dstream_id_val),
    c_apfs.APFS_TYPE.CRYPTO_STATE: (c_apfs.j_crypto_key, c_apfs.j_crypto_val),
    c_apfs.APFS_TYPE.FILE_EXTENT: (c_apfs.j_file_extent_key, c_apfs.j_file_extent_val),
    c_apfs.APFS_TYPE.DIR_REC: (c_apfs.j_drec_hashed_key, c_apfs.j_drec_val),
    c_apfs.APFS_TYPE.DIR_STATS: (c_apfs.j_dir_stats_key, c_apfs.j_dir_stats_val),
    c_apfs.APFS_TYPE.SNAP_NAME: (c_apfs.j_snap_name_key, c_apfs.j_snap_name_val),
    c_apfs.APFS_TYPE.SIBLING_MAP: (c_apfs.j_sibling_map_key, c_apfs.j_sibling_map_val),
    c_apfs.APFS_TYPE.FILE_INFO: (c_apfs.j_file_info_key, c_apfs.j_file_info_val),
}

XF_MAP = {
    c_apfs.DREC_EXT_TYPE.SIBLING_ID: c_apfs.uint64_t,
    c_apfs.INO_EXT_TYPE.SNAP_XID: c_apfs.xid_t,
    c_apfs.INO_EXT_TYPE.DELTA_TREE_OID: c_apfs.oid_t,
    c_apfs.INO_EXT_TYPE.DOCUMENT_ID: c_apfs.uint32_t,
    c_apfs.INO_EXT_TYPE.PREV_FSIZE: c_apfs.uint64_t,
    c_apfs.INO_EXT_TYPE.DSTREAM: c_apfs.j_dstream,
    c_apfs.INO_EXT_TYPE.DIR_STATS_KEY: c_apfs.j_dir_stats_val,
    c_apfs.INO_EXT_TYPE.SPARSE_BYTES: c_apfs.uint64_t,
    c_apfs.INO_EXT_TYPE.RDEV: c_apfs.uint32_t,
}



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, BinaryIO

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects import NxSuperblock

if TYPE_CHECKING:
    from uuid import UUID

    from dissect.apfs.objects.fs import FS
    from dissect.apfs.objects.keybag import ContainerKeybag

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_APFS", "CRITICAL"))


class APFS:
    """Container class for APFS operations.

    Args:
        fh: File-like object to read the APFS container from.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.fh.seek(0)

        self.sb = NxSuperblock.from_block(self, 0, self.fh.read(c_apfs.NX_DEFAULT_BLOCK_SIZE))
        self.sb.check()

        self.sbs = sorted(
            [obj for obj in self.sb.checkpoint_objects if isinstance(obj, NxSuperblock)],
            key=lambda obj: obj.xid,
            reverse=True,
        )

        # TODO: Do more accurate checkpoint traversal
        for sb in self.sbs:
            try:
                sb.check()
                sb.compare(self.sb)
            except Exception as e:
                log.debug("Skipping superblock xid=%d: %s", sb.xid, e)
                continue

            if not sb.omap.is_valid():
                log.debug("Skipping superblock xid=%d: invalid OMAP", sb.xid)
                continue

            self.sb = sb
            break

    @property
    def block_size(self) -> int:
        """The block size of the container."""
        return self.sb.block_size

    @property
    def sectors_per_block(self) -> int:
        """The number of 512-byte sectors per block."""
        return self.block_size // 512

    @property
    def block_count(self) -> int:
        """The total number of blocks in the container."""
        return self.sb.block_count

    @property
    def uuid(self) -> UUID:
        """The UUID of the container."""
        return self.sb.uuid

    @property
    def keybag(self) -> ContainerKeybag | None:
        """The container keybag, if present."""
        return self.sb.keylocker

    @property
    def volumes(self) -> list[FS]:
        """All the filesystems in the container."""
        return self.sb.filesystems

    def _read_block(self, address: int, count: int = 1) -> bytes:
        """Read a block from the container.

        Args:
            address: The block address to read.
        """
        # TODO: Fusion tier2
        self.fh.seek(address * self.block_size)
        return self.fh.read(count * self.block_size)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING
from uuid import UUID

from dissect.fve.crypto import create_cipher

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.cursor import Cursor
from dissect.apfs.exception import Error
from dissect.apfs.objects.base import Object
from dissect.apfs.objects.btree import BTree
from dissect.apfs.objects.fs import FS
from dissect.apfs.objects.keybag import ContainerKeybag
from dissect.apfs.objects.omap import ObjectMap

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.apfs.apfs import APFS
    from dissect.apfs.objects.checkpoint_map import CheckpointMap


class NxSuperblock(Object):
    """APFS NX Superblock object."""

    __type__ = c_apfs.OBJECT_TYPE_NX_SUPERBLOCK
    __struct__ = c_apfs.nx_superblock
    object: c_apfs.nx_superblock

    def check(self) -> None:
        """Check the validity of the superblock."""
        if not self.is_valid():
            raise Error("Invalid nx_superblock checksum")

        if self.type != c_apfs.OBJECT_TYPE.NX_SUPERBLOCK:
            raise Error("Invalid nx_superblock type")

        if not self.is_ephemeral:
            raise Error("Invalid nx_superblock storage type")

        if self.object.nx_magic.to_bytes(4, "big") != c_apfs.NX_MAGIC:
            raise Error(
                "Invalid nx_superblock magic "
                f"(expected {c_apfs.NX_MAGIC!r}, got {self.object.nx_magic.to_bytes(4, 'big')!r})"
            )

    def compare(self, other: NxSuperblock) -> None:
        """Compare this superblock to another superblock."""
        if self.header.o_xid < other.header.o_xid:
            raise Error("Lower xid than other superblock")

        for attr in (
            "nx_uuid",
            "nx_fusion_uuid",
            "nx_block_size",
            "nx_block_count",
            "nx_xp_desc_blocks",
            "nx_xp_data_blocks",
            "nx_xp_desc_base",
            "nx_xp_data_base",
        ):
            if getattr(self.object, attr) != getattr(other.object, attr):
                raise Error(f"Mismatch on {attr}")

    @cached_property
    def block_size(self) -> int:
        """The block size of the container."""
        return self.object.nx_block_size

    @cached_property
    def block_count(self) -> int:
        """The total number of blocks in the container."""
        return self.object.nx_block_count

    @cached_property
    def features(self) -> c_apfs.NX_FEATURE:
        """The features supported by this container."""
        return c_apfs.NX_FEATURE(self.object.nx_features)

    @cached_property
    def incompatible_features(self) -> c_apfs.NX_INCOMPAT:
        """The incompatible features supported by this container."""
        return c_apfs.NX_INCOMPAT(self.object.nx_incompatible_features)

    @cached_property
    def uuid(self) -> UUID:
        """The UUID of the container."""
        return UUID(bytes_le=self.object.nx_uuid)

    @cached_property
    def checkpoint_objects(self) -> list[CheckpointMap | NxSuperblock]:
        """All checkpoint objects in the container."""
        # TODO: Rework this a bit to be more accurate
        return list(
            _read_checkpoint_objects(self.container, self.object.nx_xp_desc_base, self.object.nx_xp_desc_blocks)
        )

    @cached_property
    def ephemeral_objects(self) -> dict[int, Object]:
        """All ephemeral objects in the container."""
        # TODO: I don't think this is correct
        return {
            obj.oid: obj
            for obj in _read_checkpoint_objects(
                self.container, self.object.nx_xp_data_base, self.object.nx_xp_data_blocks
            )
        }

    @cached_property
    def omap(self) -> ObjectMap:
        """The object map of the container."""
        return ObjectMap(self.container, self.object.nx_omap_oid)

    @cached_property
    def filesystems(self) -> list[FS]:
        """All the filesystems in the container."""
        return [FS(self.container, self.omap.lookup(oid, self.xid)) for oid in self.object.nx_fs_oid if oid != 0]

    @cached_property
    def fusion_uuid(self) -> UUID:
        """The Fusion Drive UUID."""
        return UUID(bytes_le=self.object.nx_fusion_uuid)

    @cached_property
    def keylocker(self) -> ContainerKeybag | None:
        """The container keybag, if present."""
        if self.object.nx_keylocker.pr_start_paddr == 0:
            return None

        return ContainerKeybag.from_address(
            self.container,
            self.object.nx_keylocker.pr_start_paddr,
            self.object.nx_keylocker.pr_block_count,
            cipher=create_cipher("aes-xts-128", self.object.nx_uuid * 2),
        )


def _read_checkpoint_objects(container: APFS, base: int, length: int) -> Iterator[Object]:
    """Read checkpoint objects from the container."""
    is_btree = bool(length & 0x80000000)
    num_blocks = length & 0x7FFFFFFF

    if is_btree:
        for _, value in Cursor(BTree(container, base)).walk():
            prange = c_apfs.prange(value)
            yield Object.from_address(container, prange.pr_start_paddr, prange.pr_block_count)
    else:
        for i in range(num_blocks):
            yield Object.from_address(container, base + i)



from __future__ import annotations

from dissect.apfs.c_apfs import c_apfs
from dissect.apfs.objects.base import Object


class EncryptionRollingRecoveryBlock(Object):
    """APFS Encryption Rolling Recovery Block object."""

    __type__ = c_apfs.OBJECT_TYPE_ER_RECOVERY_BLOCK
    __struct__ = c_apfs.er_recovery_block_phys
    object: c_apfs.er_recovery_block_phys
